/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#include "Commander.h"

namespace FlashMat {

Commander::Commander():
        m_i2cAddr(CELL_I2C_DEF_ADDR),
        m_pktDelayUs(0),
        m_timeoutUs(5000),
        m_waitTimeoutMs(200),
        m_numRetry(0),
        m_bcastAddresses(0),
        m_saved_i2cAddr(m_i2cAddr)
{
}

i2c_err Commander::waitForCellsToTurnOn(uint32_t msTimeout) {
    if(!m_bcastAddresses) {
        return I2C_ERR_NONE;
    }
    
    const uint8_t prevAddr = getSelectedCell();
    const uint32_t t0 = millis();
    const bool hasDeadline = (msTimeout != 0);
    
    for(uint8_t* a = m_bcastAddresses; *a != BROADCAST; ++a) {
        selectCell(*a);
        
        if(hasDeadline && millis() - t0 >= msTimeout) {
            selectCell(prevAddr);
            return I2C_ERR_TIMEOUT;
        }
        
        i2c_err res;
        while((res = ping()) != I2C_ERR_NONE) {
            if(hasDeadline && millis() - t0 >= msTimeout) {
                selectCell(prevAddr);
                return res;
            }
        }
    }
    
    selectCell(prevAddr);
    return I2C_ERR_NONE;
}

i2c_err Commander::sendRawPacket(uint8_t addr, Packet pkt, uint16_t n, const uint8_t args[], bool wait) const {
  #if SIMULATE_BROADCAST
    if(addr == BROADCAST) {
        i2c_err res = I2C_ERR_NONE;
        if(m_bcastAddresses) {
            for(const uint8_t* a = m_bcastAddresses; *a != BROADCAST; a++) {
                i2c_err r = sendRawPacket(*a, pkt, n, args, false); // never wait (wait only at the end)
                if(res == I2C_ERR_NONE) { res = r; } // notify the first error only
            }
        }
        
        if(wait) { waitCells(); }
        
        return res;
    }
  #endif
    
    i2c_err ret = I2C_ERR_UNKNOWN;
    for (uint8_t i = 0; i <= m_numRetry && !(ret == I2C_ERR_NONE || ret == I2C_ERR_TIMEOUT); i++) {
        prepareI2Ccomm(pkt);
        if(n && args) { I2C.write(args, n); }
        ret = sendI2Ccomm(addr);
    }
    
    if(wait) { waitCell(addr); }
    
    return ret;
}

i2c_err Commander::waitCells(const uint8_t addresses[]) const {
    uint8_t prevAddr = commander.getSelectedCell();
    
    const uint32_t t0 = millis();
    bool stop = false;
    for(; !stop && *addresses != BROADCAST; ++addresses) {
        uint8_t retry = 0;
        commander.selectCell(*addresses);
        
        while(commander.reqStatus().first != IDLE) { // cells must be active and idle
        //while(commander.reqStatus().first == BUSY) { // ignore disconnected cells
            if(millis() - t0 >= m_waitTimeoutMs) {
                stop = true;
                break;
            }
            
            delayUs(getRetryDelayUs(retry));
            retry++;
        }
    }
    
    commander.selectCell(prevAddr);
    
    return stop ? I2C_ERR_TIMEOUT : I2C_ERR_NONE;
}

Commander commander;

} // namespace
