/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#ifndef COMMANDER_H_
#define COMMANDER_H_

#include <Arduino.h>
#include <inttypes.h>

#include "fmat_common.h"

namespace FlashMat {

// If false, the commander never checks for a cell to become IDLE again for packets
// that perform all their calculations as soon as the packet has arrived on the cell.
// If true, the wait is performed if the caller requested it.
#define WAIT_FOR_IMMEDIATE_PKTS false

// If false, a broadcast packet is sent to the address BROADCAST.
// If true, a broadcast packet is sent singularly to all the addresses specified
// as broadcast.
#define SIMULATE_BROADCAST     true

/*
 * Utility class to be uploaded to the master MCU, that sends low-level
 * commands to the cells.
 *
 * More complex master routines can manage the generation of the image, text, etc.,
 * and delegate low-level communication to this class.
 *
 * To send commands to a FlashMat cell, first select its address with selectCell(),
 * then issue all the commands you want with the various send***() functions.
 *
 * NOTE: if you send two packets with too little a time interval between them, the
 * second packet might be dropped by the cell without notification. In some cases,
 * the packets are so close and the cell so busy that the packet is not acknowledged
 * by hardware (usually, the NAK is sent right after the I2C address).
 * Note that some errors might not get caught if you transmit in broadcast or
 * if you transmit in a non-blocking fashion.
 * This class provides an auto-retry functionality: you set the number of retry-attempts
 * (default is 0) which are automatically issued. If the problem is a timeout, no further
 * attempt is made. The timeout might be due to the cell being busy doing its calculations,
 * or it may be due to a hang in the I2C channel (which happens especially if the commander is
 * reset: the cell then desynchronizes and cannot properly recognize new transmission attempts).
 * The user might consider an I2C.resetBus() in these emergency cases.
 * The retry functionality applies to all send*** functions
 *
 * For the possible transmission errors, see the 'I2C' class (Commander simply forwards
 * I2C's errors).
 */
class Commander {
public:
    Commander(); // do not call this! use the predefined instance 'commander'

    void setup();

    void selectCell(uint8_t addr);
    uint8_t getSelectedCell() const;

    i2c_err ping() const;

    /*
     * Wait for all the listed addresses to respond to a ping.
     * All of them must respond within the deadline, or the wait will be canceled and an error will be returned.
     * If msTimeout == 0, there's no deadline (possible infinite loop).
     * msTimeout is a time interval, not a time point.
     * The deadline is for all the cells (i.e. it does not reset after one cell has been found).
     * The addresses are given through setBcastAddresses().
     */
    i2c_err waitForCellsToTurnOn(uint32_t msTimeout = 0);

    /*
     * send*():
     * Utilities to send various packets recognized by each cell.
     * Return I2C_ERR_NONE if the transmission was ok.
     *
     * 'wait' = false if the commander doesn't have to wait for the slave to complete
     * its processing. If true, the commander waits until the cell becomes IDLE again.
     */
    
    i2c_err sendSwap(SwapType type = SWAP_NOSYNC | SWAP_NOBLANK, bool wait = true) const;

    i2c_err sendStoreAddress(uint8_t i2cAddress, uint8_t rowAddress, uint8_t colAddress, bool wait = true) const;

    i2c_err sendImage4bitInChunks(const uint8_t img[FRAME_SIZE_FOR(4)], bool wait = true) const;
    i2c_err sendImageChunk4bit(uint8_t col, uint8_t row, const uint8_t img[SIZE_8x8_FOR(4)], bool wait = true) const;
    
    i2c_err sendFill(const uint8_t color[3], bool wait = true) const;

    // the font inside Text is ignored, only the id matters
    i2c_err sendTextPars(const Text& text, bool wait = true) const;

    i2c_err sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                        Fonts::FontId fontId, bool monospace,
                        uint8_t charSpacing, uint8_t lineSpacing,
                        bool wait = true) const;
    
    i2c_err sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                        Fonts::FontId fontId, bool monospace,
                        uint8_t charSpacing) const;
    
    i2c_err sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                        Fonts::FontId fontId, bool monospace) const;
    
    i2c_err sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                        Fonts::FontId fontId) const;
    
    i2c_err sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3]) const;
    
    i2c_err sendTextPars(const uint8_t color[3], bool overlay) const;
    
    i2c_err sendTextPars(const uint8_t color[3]) const;

    i2c_err sendText(const char* txt, bool wait = true) const;

    i2c_err sendCellPosition(const coord_t pos[2], bool wait = true) const;
    i2c_err sendCellPosition(coord_t x, coord_t y, bool wait = true) const;
    i2c_err sendCellPosition(Point p, bool wait = true) const;

    i2c_err sendPosition(const coord_t off[2], bool wait = true) const;
    i2c_err sendPosition(coord_t x, coord_t y, bool wait = true) const;
    i2c_err sendPosition(Point p, bool wait = true) const;

    i2c_err sendCopyBuffer(bool wait = true) const;

    i2c_err sendDrawText(bool wait = true) const;
    
    i2c_err sendDrawPixel(coord_t x, coord_t y, const uint8_t color[3], bool wait = true) const;
    i2c_err sendDrawPixel(Point p, const uint8_t color[3], bool wait = true) const;
                       
    i2c_err sendDrawLineH(coord_t x0, coord_t x1, coord_t y,
                       const uint8_t color[3], bool wait = true) const;
                       
    i2c_err sendDrawLineV(coord_t x, coord_t y0, coord_t y1,
                       const uint8_t color[3], bool wait = true) const;

    i2c_err sendDrawRect(coord_t x0, coord_t y0, coord_t x1, coord_t y1,
                      const uint8_t color[3], bool filled, bool wait = true) const;
    i2c_err sendDrawRect(Point p0, Point p1,
                      const uint8_t color[3], bool filled, bool wait = true) const;

    i2c_err sendDrawGradient(const uint8_t color1[3], const uint8_t color2[3],
                          Gradient type, bool wait = true) const;
    
    i2c_err sendDrawGradient(const uint8_t color1[3], const uint8_t color2[3],
                          Gradient type, Point topleft, Point bottomright,
                          bool wait = true) const;
    
    i2c_err sendDrawRainbow(float hue1, float hue2, float value,
                            bool wait = true) const;
    
    i2c_err sendDrawRainbow(float hue1, float hue2, float value,
                            Point topleft, Point bottomright,
                            bool wait = true) const;
    
    i2c_err sendRawPacket(uint8_t addr, Packet pkt, uint16_t n, const uint8_t args[], bool wait = true) const;

    // set the delay waited after each packet sent (to avoid sending packets too closely)
    void setPacketDelayUs(uint32_t delayUs);
    
    // timeout for I2C communication (0 = none)
    void setTimeoutUs(uint32_t us);
    
    // number of auto-retry attempts when the packet was not properly received
    void setNumRetry(uint8_t num);
    
    // timeout when waiting for a cell to become idle
    void setWaitTimeoutMs(uint16_t ms);

    // May return invalid values (e.g. 0xFF) if the cell is not there or any error happens,
    // so you should check for a 'default' or 'else' case when dealing with the returned values.
    std::pair<CellStatus, PacketStatus> reqStatus() const;
    
    // BROADCAST-terminated array of addresses, used for simulated broadcast and for waiting.
    bool setBcastAddresses(const uint8_t addresses[]);
    
    i2c_err waitCurrentCell() const; // wait the current addressed cell, or the waitAddresses, if the current cell is BROADCAST
    i2c_err waitCell(uint8_t address) const; // wait for the given address to become idle
    i2c_err waitCells(const uint8_t addresses[]) const; // wait for the given addresses (BROADCAST-terminated) to become idle
    i2c_err waitCells() const; // wait for the saved waitAddresses (BROADCAST-terminated) to become idle
    
private:
    uint8_t m_i2cAddr;
    
    uint32_t m_pktDelayUs;
    uint32_t m_timeoutUs;
    uint32_t m_waitTimeoutMs;
    uint8_t m_numRetry;
    
    uint8_t* m_bcastAddresses; // BROADCAST-terminated array of addresses

    void initPins() const;
    void initI2C() const;
    
    i2c_err _hlpSendTextPars(uint8_t numPars, const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                            Fonts::FontId fontId, bool monospace,
                            uint8_t charSpacing, uint8_t lineSpacing,
                            bool wait = true) const;

    // For each retry, set a delay (us) to be waited before retrying,
    // to avoid slowing down the target cell
    uint16_t getRetryDelayUs(uint8_t retry_num) const;
    
    void prepareI2Ccomm(Packet packet) const;
    i2c_err sendI2Ccomm(uint8_t addr) const;
    
    uint8_t m_saved_i2cAddr;
};

extern Commander commander;


///////////////////////////////////////////////////////
// Inline-able implementation
///////////////////////////////////////////////////////

inline
void Commander::setPacketDelayUs(uint32_t delayUs) {
    m_pktDelayUs = delayUs;
}

inline
void Commander::setTimeoutUs(uint32_t us) {
    m_timeoutUs = us;
}
    
inline
void Commander::setNumRetry(uint8_t num) {
    m_numRetry = (num == 255 ? 254 : num); // avoid endless loop in the current implementation of send***() functions
}

inline
void Commander::setWaitTimeoutMs(uint16_t ms) {
    m_waitTimeoutMs = ms;
}

inline
i2c_err Commander::ping() const {
    if(m_i2cAddr != BROADCAST) {
        return sendRawPacket(m_i2cAddr, PKT_PING, 0, 0, false);
    } else {
        // Real Broadcast, not the simulated one
        i2c_err ret = I2C_ERR_UNKNOWN;
        for (uint8_t i = 0; i <= m_numRetry && !(ret == I2C_ERR_NONE || ret == I2C_ERR_TIMEOUT); i++) {
            prepareI2Ccomm(PKT_PING);
            ret = sendI2Ccomm(BROADCAST);
        }
        return ret;
    }
}

inline
i2c_err Commander::sendStoreAddress(uint8_t i2cAddress, uint8_t rowAddress, uint8_t colAddress, bool wait) const {
    if(m_i2cAddr != BROADCAST) {
        const uint8_t N = 4;
        uint8_t args[N] = {(uint8_t)(~PKT_STORE_ADDRESS), // extra guarantee that no error-induced address change can happen
                           i2cAddress,
                           rowAddress,
                           colAddress
                           };
        i2c_err err = sendRawPacket(m_i2cAddr, PKT_STORE_ADDRESS, N, args, false); // no wait: we must wait for the new address!
        if(wait && err == I2C_ERR_NONE) {
            return waitCell(i2cAddress);
        }
        return err;
    } else {
        // Real Broadcast, not the simulated one
        i2c_err ret = I2C_ERR_UNKNOWN;
        for (uint8_t i = 0; i <= m_numRetry && !(ret == I2C_ERR_NONE || ret == I2C_ERR_TIMEOUT); i++) {
            prepareI2Ccomm(PKT_STORE_ADDRESS);
            I2C.write((uint8_t)(~PKT_STORE_ADDRESS)); // extra guarantee that no error-induced address change can happen
            I2C.write(i2cAddress);
            I2C.write(rowAddress);
            I2C.write(colAddress);
            ret = sendI2Ccomm(BROADCAST);
        }
        return ret;
    }
}

inline
i2c_err Commander::sendSwap(SwapType type, bool wait) const {
    const uint8_t t = type;
    return sendRawPacket(m_i2cAddr, PKT_SWAP, 1, &t, wait);
}

inline
i2c_err Commander::sendImageChunk4bit(uint8_t col, uint8_t row, const uint8_t img[SIZE_8x8_FOR(4)], bool wait) const {
    const uint8_t N = 2 + SIZE_8x8_FOR(4);
    uint8_t args[N];
    args[0] = col;
    args[1] = row;
    memcpy(args + 2, img, SIZE_8x8_FOR(4));
    return sendRawPacket(m_i2cAddr, PKT_IMG_4bit_CHUNK, N, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendImage4bitInChunks(const uint8_t img[FRAME_SIZE_FOR(4)], bool wait) const {
    i2c_err ret = I2C_ERR_NONE;
    
    uint8_t chunk[SIZE_8x8_FOR(4)];
    for(uint8_t c = 0; c < MATRIX_COLS / 8; c++) {
        for(uint8_t r = 0; r < MATRIX_ROWS / 8; r++) {
            ImgChunker::getChunk4bit(chunk, img, c, r);
            i2c_err r2 = sendImageChunk4bit(c, r, chunk, (c == MATRIX_COLS / 8 - 1 && r == MATRIX_ROWS / 8 - 1) ? wait : true); // always wait if it's not the last one
            
            if(ret == I2C_ERR_NONE) {
                // notify the first error only
                ret = r2;
            }
        }
    }
    
    return ret;
}

inline
i2c_err Commander::sendFill(const uint8_t color[3], bool wait) const {
    const uint8_t N = 3;
    uint8_t args[N];
    memcpy(args, color, 3);
    return sendRawPacket(m_i2cAddr, PKT_FILL, N, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::_hlpSendTextPars(uint8_t numPars,
                                    const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
                                    Fonts::FontId fontId, bool monospace,
                                    uint8_t charSpacing, uint8_t lineSpacing,
                                    bool wait) const
{
    //assert(numPars >= 1 && numPars <= 7);
    
    const uint8_t N = 3 + 1 + 3 + 1 + 1 + 1 + 1;
    uint8_t actualN = 0;
    uint8_t args[N];
    
    if(numPars >= 1) { actualN =  3; memcpy(args     , color          , 3); }
    if(numPars >= 2) { actualN =  4;        args[3]  = overlay;             }
    if(numPars >= 3) { actualN =  7; memcpy(args +  4, bgColor        , 3); }
    if(numPars >= 4) { actualN =  8;        args[7]  = (uint8_t)fontId;     }
    if(numPars >= 5) { actualN =  9;        args[8]  = monospace;           }
    if(numPars >= 6) { actualN = 10;        args[9]  = charSpacing;         }
    if(numPars >= 7) { actualN = 11;        args[10] = lineSpacing;         }
    return sendRawPacket(m_i2cAddr, PKT_TEXT_PARS, actualN, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendTextPars(const Text& text, bool wait) const {
    return sendTextPars(text.color, text.overlay, text.bgColor,
                        text.fontId, text.monospace, text.charSpacing, text.lineSpacing,
                        wait);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
        Fonts::FontId fontId, bool monospace, uint8_t charSpacing, uint8_t lineSpacing, bool wait) const {
    
    return _hlpSendTextPars(7, color, overlay, bgColor,
                            fontId, monospace, charSpacing, lineSpacing,
                            wait);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
        Fonts::FontId fontId, bool monospace, uint8_t charSpacing) const {
    
    return _hlpSendTextPars(6, color, overlay, bgColor, fontId, monospace, charSpacing, 0);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
        Fonts::FontId fontId, bool monospace) const {
    
    return _hlpSendTextPars(5, color, overlay, bgColor, fontId, monospace, 0, 0);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3],
        Fonts::FontId fontId) const {
    
    return _hlpSendTextPars(4, color, overlay, bgColor, fontId, 0, 0, 0);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay, const uint8_t bgColor[3]) const {
    
    return _hlpSendTextPars(3, color, overlay, bgColor, (Fonts::FontId)0, 0, 0, 0);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3], bool overlay) const {
    
    return _hlpSendTextPars(2, color, overlay, 0, (Fonts::FontId)0, 0, 0, 0);
}

inline
i2c_err Commander::sendTextPars(const uint8_t color[3]) const {
    
    return _hlpSendTextPars(1, color, 0, 0, (Fonts::FontId)0, 0, 0, 0);
}

inline
i2c_err Commander::sendText(const char* txt, bool wait) const {
    uint16_t len = strlen(txt);

    i2c_err res = I2C_ERR_UNKNOWN;
    for (uint16_t p = 0, n = 0; p < len; p += TEXT_PACKET_MAX_SIZE, ++n) {
        for (uint8_t i = 0; i <= m_numRetry && !(res == I2C_ERR_NONE || res == I2C_ERR_TIMEOUT); i++) {
            uint8_t args[TEXT_PACKET_MAX_SIZE + 1];
            args[0] = n;

            bool lastPkt = (p + (uint16_t)TEXT_PACKET_MAX_SIZE >= len);

            uint16_t chunkLen = lastPkt ? len - p : TEXT_PACKET_MAX_SIZE;
            memcpy(args + 1, txt + p, chunkLen);

            i2c_err r = sendRawPacket(m_i2cAddr, PKT_TEXT, chunkLen + 1, args, (lastPkt ? wait : true));
            
            if (res != I2C_ERR_NONE) { res = r; } // report only the first error
        }
    }
    return res;
}

inline
i2c_err Commander::sendCellPosition(Point p, bool wait) const {
    return sendCellPosition(p.x, p.y, wait);
}

inline
i2c_err Commander::sendCellPosition(const coord_t pos[2], bool wait) const {
    return sendCellPosition(pos[X], pos[Y], wait);
}

inline
i2c_err Commander::sendCellPosition(coord_t x, coord_t y, bool wait) const {
    // broadcast is allowed, but the intended use is to give each cell
    // its position inside the matrix, so broadcast should be avoided
    // NOTE: position is still given in pixels, not in cells
    // (the second cell is at [MATRIX_COLS,0], not at [1,0])

    const uint8_t N = 2 * sizeof(coord_t);
    uint8_t args[N];
    serialize<coord_t>(x, args);
    serialize<coord_t>(y, args + sizeof(coord_t));
    return sendRawPacket(m_i2cAddr, PKT_CELL_POSITION, N, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendPosition(Point p, bool wait) const {
    return sendPosition(p.x, p.y, wait);
}

inline
i2c_err Commander::sendPosition(const coord_t pos[2], bool wait) const {
    return sendPosition(pos[X], pos[Y], wait);
}

inline
i2c_err Commander::sendPosition(coord_t x, coord_t y, bool wait) const {
    const uint8_t N = 2 * sizeof(coord_t);
    uint8_t args[N];
    serialize<coord_t>(x, args);
    serialize<coord_t>(y, args + sizeof(coord_t));
    return sendRawPacket(m_i2cAddr, PKT_TEXT_POSITION, N, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendCopyBuffer(bool wait) const {
    return sendRawPacket(m_i2cAddr, PKT_COPY_BUFFER, 0, 0, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendDrawText(bool wait) const {
    return sendRawPacket(m_i2cAddr, PKT_DRAW_TEXT, 0, 0, wait);
}

inline
i2c_err Commander::sendDrawPixel(Point p, const uint8_t color[3], bool wait) const {
    return sendDrawPixel(p.x, p.y, color, wait);
}

inline
i2c_err Commander::sendDrawPixel(coord_t x, coord_t y, const uint8_t color[3], bool wait) const {
    const uint8_t N = 2*sizeof(coord_t) + 3;
    uint8_t args[N];
    serialize<coord_t>(x, args);
    serialize<coord_t>(y, args + sizeof(coord_t));
    memcpy(args + 2*sizeof(coord_t), color, 3);
    return sendRawPacket(m_i2cAddr, PKT_DRAW_PIXEL, N, args, WAIT_FOR_IMMEDIATE_PKTS ? wait : false);
}

inline
i2c_err Commander::sendDrawLineH(coord_t x0, coord_t x1, coord_t y,
                              const uint8_t color[3], bool wait) const
{
    const uint8_t N = 3*sizeof(coord_t) + 3;
    uint8_t args[N];
    serialize<coord_t>(x0, args);
    serialize<coord_t>(x1, args + sizeof(coord_t));
    serialize<coord_t>(y , args + 2*sizeof(coord_t));
    memcpy(args + 3*sizeof(coord_t), color, 3);
    return sendRawPacket(m_i2cAddr, PKT_DRAW_LINE_H, N, args, wait);
}

inline
i2c_err Commander::sendDrawLineV(coord_t x, coord_t y0, coord_t y1,
                              const uint8_t color[3], bool wait) const
{
    const uint8_t N = 3*sizeof(coord_t) + 3;
    uint8_t args[N];
    serialize<coord_t>(x , args);
    serialize<coord_t>(y0, args + sizeof(coord_t));
    serialize<coord_t>(y1, args + 2*sizeof(coord_t));
    memcpy(args + 3*sizeof(coord_t), color, 3);
    return sendRawPacket(m_i2cAddr, PKT_DRAW_LINE_V, N, args, wait);
}

inline
i2c_err Commander::sendDrawRect(Point p0, Point p1,
                             const uint8_t color[3], bool filled, bool wait) const
{
    return sendDrawRect(p0.x, p0.y, p1.x, p1.y, color, filled, wait);
}

inline
i2c_err Commander::sendDrawRect(coord_t x0, coord_t y0, coord_t x1, coord_t y1,
                             const uint8_t color[3], bool filled, bool wait) const
{
    const uint8_t N = 4*sizeof(coord_t) + 3 + 1;
    uint8_t args[N];
    serialize<coord_t>(x0, args);
    serialize<coord_t>(y0, args + sizeof(coord_t));
    serialize<coord_t>(x1, args + 2*sizeof(coord_t));
    serialize<coord_t>(y1, args + 3*sizeof(coord_t));
    memcpy(args + 4*sizeof(coord_t), color, 3);
    args[4*sizeof(coord_t) + 3] = filled;
    return sendRawPacket(m_i2cAddr, PKT_DRAW_RECT, N, args, wait);
}

inline
i2c_err Commander::sendDrawGradient(const uint8_t color1[3],
                                 const uint8_t color2[3],
                                 Gradient type, bool wait) const
{
    const uint8_t N = 2*3 + 1;
    uint8_t args[N];
    memcpy(args    , color1, 3);
    memcpy(args + 3, color2, 3);
    args[2*3] = (uint8_t) type;
    return sendRawPacket(m_i2cAddr, PKT_DRAW_GRADIENT, N, args, wait);
}

inline
i2c_err Commander::sendDrawGradient(const uint8_t color1[3],
                                 const uint8_t color2[3],
                                 Gradient type,
                                 Point topleft, Point bottomright,
                                 bool wait) const
{
    const uint8_t N = 2*3 + 1 + 4*sizeof(coord_t);
    uint8_t args[N];
    memcpy(args    , color1, 3);
    memcpy(args + 3, color2, 3);
    args[2*3] = (uint8_t) type;
    serialize<coord_t>(topleft.x, args + 2*3 + 1);
    serialize<coord_t>(topleft.y, args + 2*3 + 1 + sizeof(coord_t));
    serialize<coord_t>(bottomright.x, args + 2*3 + 1 + 2*sizeof(coord_t));
    serialize<coord_t>(bottomright.y, args + 2*3 + 1 + 3*sizeof(coord_t));
    return sendRawPacket(m_i2cAddr, PKT_DRAW_GRADIENT, N, args, wait);
}

inline
i2c_err Commander::sendDrawRainbow(float hue1, float hue2, float value, bool wait) const
{
    const uint8_t N = 3*sizeof(float);
    uint8_t args[N];
    serialize<float>(hue1, args);
    serialize<float>(hue2, args + sizeof(float));
    serialize<float>(value, args + 2*sizeof(float));
    return sendRawPacket(m_i2cAddr, PKT_DRAW_RAINBOW, N, args, wait);
}

inline
i2c_err Commander::sendDrawRainbow(float hue1, float hue2, float value,
                                    Point topleft, Point bottomright,
                                    bool wait) const
{
    const uint8_t N = 3*sizeof(float) + 4*sizeof(coord_t);
    uint8_t args[N];
    serialize<float>(hue1, args);
    serialize<float>(hue2, args + sizeof(float));
    serialize<float>(value, args + 2*sizeof(float));
    serialize<coord_t>(topleft.x, args + 3*sizeof(float));
    serialize<coord_t>(topleft.y, args + 3*sizeof(float) + sizeof(coord_t));
    serialize<coord_t>(bottomright.x, args + 3*sizeof(float) + 2*sizeof(coord_t));
    serialize<coord_t>(bottomright.y, args + 3*sizeof(float) + 3*sizeof(coord_t));
    return sendRawPacket(m_i2cAddr, PKT_DRAW_RAINBOW, N, args, wait);
}

inline
std::pair<CellStatus, PacketStatus> Commander::reqStatus() const {
    uint8_t* data;
    i2c_err ret = I2C_ERR_UNKNOWN;
    for (uint8_t i = 0; i <= m_numRetry && !(ret >= 0 || ret == I2C_ERR_TIMEOUT); i++) {
        ret = I2C.requestFrom(m_i2cAddr, data, 2);
    }
    
    if(ret >= 2 && data) {
        return std::pair<CellStatus, PacketStatus>((CellStatus)data[0], (PacketStatus)data[1]);
    } else {
        return std::pair<CellStatus, PacketStatus>((CellStatus)0xFF, (PacketStatus)0xFF);
    }
}

inline
void Commander::prepareI2Ccomm(Packet packet) const {
    while (!I2C.begin()); // wait for I2C to be ready
    I2C.write(packet);
}

inline
i2c_err Commander::sendI2Ccomm(uint8_t addr) const {
    i2c_err res = I2C.send(addr, true);
    if(res != I2C_ERR_NONE) {
        return res;
    }
    delayUs(m_pktDelayUs);
    return I2C_ERR_NONE;
}

inline
void Commander::selectCell(uint8_t addr) {
    if(I2C_Class::isValidAddress(addr)) {
        this->m_i2cAddr = addr;
    }
}

inline
uint8_t Commander::getSelectedCell() const {
    return m_i2cAddr;
}

inline
uint16_t Commander::getRetryDelayUs(uint8_t retry_num) const {
    switch(retry_num) {
        case 0:
            return  250;
        case 1:
            return  500;
        case 2:
            return 1000;
        case 3:
            return 2000;
        case 4:
            return 4000;
        default:
            return 8000;
    }
}

inline
bool Commander::setBcastAddresses(const uint8_t addresses[]) {
    if(!addresses) {
        free(m_bcastAddresses);
        m_bcastAddresses = 0;
        return true;
    } else {
        uint8_t len = 0;
        for(const uint8_t* p = addresses; *p++ != BROADCAST; len++);
        m_bcastAddresses = static_cast<uint8_t*>(realloc(m_bcastAddresses, len + 1));
        
        if(!m_bcastAddresses) {
            return false;
        } else {
            for(uint8_t* p = m_bcastAddresses; (*p++ = *addresses++) != BROADCAST;);
            return true;
        }
    }
}

inline
i2c_err Commander::waitCurrentCell() const {
    if(m_i2cAddr == BROADCAST) {
        return waitCells();
    } else {
        return waitCell(m_i2cAddr);
    }
}

inline
i2c_err Commander::waitCell(uint8_t address) const {
    uint8_t addrs[2] = {address, BROADCAST};
    return waitCells(addrs);
}

inline
i2c_err Commander::waitCells() const {
    if(m_bcastAddresses) {
        return waitCells(m_bcastAddresses);
    } else {
        return I2C_ERR_NONE;
    }
}

inline
void Commander::setup() {
    initPins();
    initI2C();
    
    I2C.resetBus();
}

inline
void Commander::initPins() const {
    // nothing to do, SDA and SCL are handled by the I2C library
}

inline
void Commander::initI2C() const {
    I2C.setFreq(FM_I2C_FREQ);
    I2C.setBufferLength(0, 0, 2, FM_I2C_BUFFER_SIZE); // buffer of length 2: to receive the cell/packet status
    
    // this is a master only: not reachable by other masters
    I2C.setAddress(0);
    I2C.receiveGeneralCalls(false);

    I2C.enablePullups(); // always enable them on the commander. Only the cell has them external

    I2C.enable();
}

} // namespace

#endif /* COMMANDER_H_ */
