/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#ifndef FONT_H_
#define FONT_H_

#include <avr/pgmspace.h>
#include <inttypes.h>

/*
 * Represents characters of maximum NxM size.
 * Each character has a bitmask (NxM) to represent foreground and background
 * (stored in bytes, row-by-row).
 * The font as a whole has a height (the height of the tallest character)
 * and an y-offset. Each character has an x-offset and a width.
 * With these 4 numbers you can define a "window" where the character is shown;
 * such window is used to allow nice spacing between variable-sized characters.
 * You can also ignore these values and show each character as an 8x8 matrix,
 * i.e. in a monospace fashion.
 * NOTE: pixels outside this "window" are still drawn. The window is only used
 * for placing and spacing the characters.
 * For example, by defining offset and size as x0=8 w=0, the character will
 * be "shifted" 8 pixels to the left (hence overlapping with the previous one),
 * and take up no additional space.
 *
 * For monospace fonts, you can define a monoHeight and monoWidth parameters,
 * which define the size of the character.
 * Each font can be used in monospace or non-monospace mode.
 * In non-monospace mode: x0,y0,height,width define the position of the char on
 * the screen and the advance of the cursor (plus lineSpacing and charSpacing,
 * defined outside the font).
 * In monospace mode: x0,y0 are assumed to be 0. monoHeight and monoWidth define
 * the advance of the cursor (plus lineSpacing and charSpacing).
 * In both monospace and non-monospace modes, pixels outside the window are still
 * drawn (so, they should usually be 0's).
 */
/*
 * Font settings are stored in program memory because they alone would require
 * more RAM than is available on the MCU.
 * So, the font cannot be changed dinamically; you can only switch between
 * predefined fonts.
 */
/*
 * NOTE to create new fonts:
 * create a copy of one font file (e.g. "font_fm_8x8.cpp"), rename its object (e.g. "font_fm_8x8")
 * to whatever you like and change its settings.
 * (Note: the newline '\n' character and the null '\0' character are treated in a special
 * manner, and are never drawn, so they should always have that meaning in every custom
 * font. Spaces, tabs and control characters are treated as glyphs, i.e. they are printed
 * like any normal letter would, according to their mask).
 *
 * On the cell device: Fonts.cpp must be updated with the reference to all available
 * fonts, and it must be queried at runtime for the selected font.
 *
 * On the master device: you can send a packet to the cells, telling them
 * which font to use (by giving the font's number). The master needn't have a copy
 * of Fonts.cpp, unless it wants to display by itself some text and then send
 * to the cell the pre-rendered image.
 */

namespace FlashMat {

class Font {

public:
	Font(const uint8_t* chars/*PROGMEM*/);

	uint8_t minChar() const; // >= 1
	uint8_t maxChar() const; // in [1,255]; 0 if no character exist (a "no.op" font)

	int8_t y0() const;
	uint8_t height() const;

	uint8_t monoHeight() const;
	uint8_t monoWidth() const;

	uint8_t maskHeight() const;
	uint8_t maskWidth() const;
    uint8_t maskBytesPerChar() const;

	// NOTE: calls x0(0) and width(0) shouldn't be made: they will
	// return y0() and height().
	// getMask(0) and getMaskRow(0) shouldn't be made: their return value
	// is undefined (though as of now the implementation returns all 0's)

	int8_t x0(uint8_t c) const; // Returns the x-offset of the specified character
	uint8_t width(uint8_t c) const; // Returns the width of the specified character

	// Fills 'mask' with the bitmask of the character
	void getMask(uint8_t c, uint8_t mask[/*maskBytesPerChar()*/]) const;


private:
    Font(); // unused
    
	// Used to access the 'chars' matrix with proper number of rows/columns
	uint16_t chr(uint8_t charNum, uint8_t byteNum) const;
	uint8_t bytesPerChar() const;
    
    bool charExists(uint8_t c) const;

	/*
	 * Each row of this matrix is the description of one character:
	 * + x0 is the horizontal offset:
	 *    0 = the char starts on the first column
	 *    1 = it starts on the second column
	 *   -1 = one blank column is "added" to the left ("added" means that it's handled
	 *       by the class that actually draws the characters)
	 * + width: the advance of the cursor, to find the position of the next
	 *    char (+ charSpacing)
	 * + mask: N bytes describing the char (1's = char, 0's = background)
	 *
	 * The null char (entry 0 of this array) is used to hold these values:
	 * + y0 offset
	 * + height
	 * + monoHeight
	 * + monoWidth
	 * + numChars of this font (127 = 0-127; 255 = 0-255)
	 * + maskH (1 = 8 pixels per column, 2 = 16 pixels, 3 = 24 pixels,...)
	 * + maskW (1 = 8 pixels per row, 2 = 16 pixels, 3 = 24 pixels,...)
	 *
	 * The null char is never shown, so it needn't have its entry here.
	 *
	 * The font height is the max height of all characters. It's used to correctly
	 * space text lines (new-line y-coord = old-line y-coord + height + spacing).
	 *
	 * y0 is the vertical offset:
	 * 0 = the chars start on the first row
	 * 1 = they start on the second row
	 * -1 = one blank row is "added" at the top ("added" means that it's handled
	 * by the class that actually draws the characters)
	 *
	 * When the font is shown as monospaced, x0 and y0 are assumed to be 0 for
	 * all characters, and monoHeight and monoWidth are used instead
	 * of height and width, for all characters (charSpacing and lineSpacing
	 * are ignored)
	 */
	const uint8_t* _chars/*PROGMEM*/;

	uint8_t _minChar;
	uint8_t _maxChar;
	uint8_t _maskBytesPerChar;
	int8_t _y0;
	uint8_t _height;
	uint8_t _monoHeight;
	uint8_t _monoWidth;
	uint8_t _maskHeight;
	uint8_t _maskWidth;
};

inline uint16_t Font::chr(uint8_t charNum, uint8_t byteNum) const{
	return byteNum + (charNum - minChar() + 1) * bytesPerChar();
}

inline uint8_t Font::minChar() const {
	return _minChar;
}

inline uint8_t Font::maxChar() const {
    return _maxChar;
}


inline int8_t Font::y0() const {
	return _y0;
}

inline uint8_t Font::height() const {
	return _height;
}

inline uint8_t Font::monoHeight() const {
	return _monoHeight;
}

inline uint8_t Font::monoWidth() const {
	return _monoWidth;
}

inline uint8_t Font::maskHeight() const {
	return _maskHeight;
}

inline uint8_t Font::maskWidth() const {
	return _maskWidth;
}

inline uint8_t Font::maskBytesPerChar() const {
    return _maskBytesPerChar;
}

inline uint8_t Font::bytesPerChar() const {
    return maskBytesPerChar() + 2; // mask + 1byte x0 + 1byte width
}

inline bool Font::charExists(uint8_t c) const {
    return c >= minChar() && c <= maxChar() && c != 0;
}

}

#endif /* FONT_H_ */
