/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "I2C.h"

#include <math.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <compat/twi.h>

#include "pins_arduino.h"

I2C_Class::I2C_Class() :
        _pullups(true), //
		_state(I2C_STATE_READY), //
		_slarw(0x00), //
		_sendStop(true), //
		_inRepStart(false), //
		_error(TW_NO_INFO), //
		_gCall(false), //
		_slaveOnReceive(0), //
		_slaveOnRequest(0), //
		MAX_SLAVE_RX_BUFFER_LENGTH(0), //
		MAX_SLAVE_TX_BUFFER_LENGTH(0), //
		MAX_MASTER_RX_BUFFER_LENGTH(0), //
		MAX_MASTER_TX_BUFFER_LENGTH(0), //
		_slaveRxBuffer(0), //
		_slaveRxBufferIndex(0), //
		_slaveTxBuffer(0), //
		_slaveTxBufferIndex(0), //
		_slaveTxBufferLength(0), //
		_masterRxBuffer(0), //
		_masterRxBufferIndex(0), //
		_masterRxBufferLength(0), //
		_masterTxBuffer(0), //
		_masterTxBufferIndex(0), //
		_masterTxBufferLength(0) //
{
	// own address 0 and disable general calls: the device should not respond
	// to anything until proper settings are made (and respond to nothing if
	// TWI is enabled before setting a proper address)
	TWAR = 0x00;
}

/*
 * Default settings. If you don't set frequency and buffer lengths explicitly,
 * you need to call this function, or the library won't work.
 *
 * Do not call this while the channel is enabled.
 */
void I2C_Class::defaultSettings() {
	setFreq(100000L);
	setBufferLength(32, 32, 32, 32);
	enablePullups();
	TWAR = 0x00; // address 0x00 (invalid) and no general calls
}

/*
 * Set the frequency.
 *
 * Returns true if it is feasible; if not, the frequency is left unchanged.
 *
 * Do not call this while the channel is enabled.
 */
bool I2C_Class::setFreq(uint32_t frequency) {
	// Initialize twi prescaler
    // TWPS1 TWPS0   prescaler
    //   0     0        /1
    //   0     1        /4
    //   1     0        /16
    //   1     1        /64
	BSET0(TWSR, TWPS0);
	BSET0(TWSR, TWPS1);

	// Initialize bitrate
	uint8_t twbr_value = (uint8_t) (((F_CPU / frequency) - 16) / 2);
	/* twi bit rate formula from atmega328 manual pg 216
	 SCL_Frequency = CPU_Clock_Frequency / (16 + (2 * TWBR)*Prescaler_Value)
	 note: TWBR should be 10 or higher for master mode
	 It is 72 for a 16mhz Wiring board with 100kHz TWI */
	/*
	 * The frequency is limited by the bus capacitance.
	 * Rise/fall time is proportional to tau=RC, so the pullup resistor must be
	 * lower as the bus capacitance increases. For 400kHz (Fast mode twi) R can
	 * be as low as 1.5k~4.7k
	 *
	 * Moreover, a slow slave can keep the bus busy until it has finished processing
	 * its data, so the actual frequency may be lower than the one used by the master
	 */

	if (twbr_value > 0) {
		TWBR = twbr_value;
		return true;
	} else {
		return false;
	}
}

/*
 * Set the length for each of the four buffers.
 *
 * Returns false if there's not enough memory.
 * NOTE: if the memory is not enough, all buffers are deallocated.
 * Do not call this while the channel is enabled or some tx message bytes
 * have already been buffered.
 */
bool I2C_Class::setBufferLength(uint8_t slaveRxLen, uint8_t slaveTxLen,
		uint8_t masterRxLen, uint8_t masterTxLen) {
	// allocate buffers:
	// reallocate only if size is different; then check that the allocation
	// was successful, otherwise free all the memory and return false

	if (MAX_SLAVE_RX_BUFFER_LENGTH != slaveRxLen) {
		MAX_SLAVE_RX_BUFFER_LENGTH = slaveRxLen;
		free(_slaveRxBuffer);
		_slaveRxBuffer = 0;
		_slaveRxBufferIndex = 0;
	}

	if (MAX_SLAVE_TX_BUFFER_LENGTH != slaveTxLen) {
		MAX_SLAVE_TX_BUFFER_LENGTH = slaveTxLen;
		free(_slaveTxBuffer);
		_slaveTxBuffer = 0;
		_slaveTxBufferIndex = 0;
		_slaveTxBufferLength = 0;
	}

	if (MAX_MASTER_RX_BUFFER_LENGTH != masterRxLen) {
		MAX_MASTER_RX_BUFFER_LENGTH = masterRxLen;
		free(_masterRxBuffer);
		_masterRxBuffer = 0;
		_masterRxBufferIndex = 0;
		_masterRxBufferLength = 0;
	}

	if (MAX_MASTER_TX_BUFFER_LENGTH != masterTxLen) {
		MAX_MASTER_TX_BUFFER_LENGTH = masterTxLen;
		free(_masterTxBuffer);
		_masterTxBuffer = 0;
		_masterTxBufferIndex = 0;
		_masterTxBufferLength = 0;
	}

	if (!_slaveRxBuffer && slaveRxLen > 0 // slaveRx to be allocated
			) {
		_slaveRxBuffer = (uint8_t*) malloc(slaveRxLen);
	}
	if ((_slaveRxBuffer || slaveRxLen == 0) && // slaveRx unused or correctly allocated
			(!_slaveTxBuffer && slaveTxLen > 0) // slaveTx to be allocated
			) {
		_slaveTxBuffer = (uint8_t*) malloc(slaveTxLen);
	}
	if ((_slaveRxBuffer || slaveRxLen == 0) && // slaveRx unused or correctly allocated
			(_slaveTxBuffer || slaveTxLen == 0) && // slaveTx unused or correctly allocated
			(!_masterRxBuffer && masterRxLen > 0) // masterRx to be allocated
			) {
		_masterRxBuffer = (uint8_t*) malloc(masterRxLen);
	}
	if ((_slaveRxBuffer || slaveRxLen == 0) && // slaveRx unused or correctly allocated
			(_slaveTxBuffer || slaveTxLen == 0) && // slaveTx unused or correctly allocated
			(_masterRxBuffer || masterRxLen == 0) && // masterRx unused or correctly allocated
			(!_masterTxBuffer && masterTxLen > 0) // masterTx to be allocated
			) {
		_masterTxBuffer = (uint8_t*) malloc(masterTxLen);
	}

	if ((!_slaveRxBuffer && slaveRxLen > 0) || // slaveRx needed and not correctly allocated
			(!_slaveTxBuffer && slaveTxLen > 0) || // slaveTx needed and not correctly allocated
			(!_masterRxBuffer && masterRxLen > 0) || // masterRx needed and not correctly allocated
			(!_masterTxBuffer && masterTxLen > 0) // masterTx needed and not correctly allocated
			) {
		MAX_SLAVE_RX_BUFFER_LENGTH = 0;
		MAX_SLAVE_TX_BUFFER_LENGTH = 0;
		MAX_MASTER_RX_BUFFER_LENGTH = 0;
		MAX_MASTER_TX_BUFFER_LENGTH = 0;
		free(_slaveRxBuffer);
		free(_slaveTxBuffer);
		free(_masterRxBuffer);
		free(_masterTxBuffer);
		_slaveRxBuffer = 0;
		_slaveTxBuffer = 0;
		_masterRxBuffer = 0;
		_masterTxBuffer = 0;
		return false;
	}

	return true;
}

/*
 * Set this device's address.
 * 8~119 are valid addresses.
 * 0 can be used for "address-less" devices (no one can transmit to them,
 *  except with general call, if enabled).
 * 1~7 and 120~127 are reserved.
 * 128~255 are not addresses (there are only 7 bits)
 *
 * Return true if the address is allowed; otherwise the address is left
 * unchanged.
 *
 * Do not call this while the channel is enabled.
 */
bool I2C_Class::setAddress(uint8_t address) {
	// Address 0x00: reserved to general call, but allowed to make a device
	// "address-less", so that it can receive no data, or only general call data
	// Addresses 1~7 and 120~127: reserved by the I2C specs, but allowed here
	// just in case someone needs them.

	if (!isValidAddress(address, true))
		return false;

	// TWAR[7:1] = address
	// TWAR[0] = TWGCE (general call enable)
	TWAR &= 0x01; // delete old address
	TWAR |= (address << 1); // write new one

	return true;
}

/*
 * Master sends a request of 'quantity' bytes to a slave, and waits for
 * the response.
 *
 * Returns  >=0 : the number of bytes read
 *          or I2C_ERR_ARB_LOST
 *          or I2C_ERR_ADDR_NACK
 *          or I2C_ERR_BUS
 *          or I2C_ERR_UNKNOWN
 *          or ERR_RESERVED_ADDR
 *
 * NOTE: the master always asks for 'quantity' bytes; if the slave doesn't provide
 * enough data, garbage is read. To avoid this you must implement some software
 * strategy (see the class main comment).
 *
 * 'data' is a byte* provided by the caller; it is a pointer, which is changed by
 * the function to point to the location where the received data is stored.
 * The caller must read and/or copy from there before issuing new requests, or
 * the data will be overwritten.
 */
i2c_err I2C_Class::requestFrom(uint8_t address, uint8_t*& data,
		uint8_t quantity, bool sendStop) {
	
	if(!isValidAddress(address, false)) {
        data = 0;
		return ERR_RESERVED_ADDR;
	}
	
	// clamp to buffer length
	if (quantity > MAX_MASTER_RX_BUFFER_LENGTH) {
		quantity = MAX_MASTER_RX_BUFFER_LENGTH;
	}

	// perform blocking read into buffer
    i2c_err res = doMasterRequestTx(address, quantity, sendStop);
    
    data = (res >= 0) ? _masterRxBuffer : 0;
    
    return res;
}

/*
 * Starts the buffering of a new transmission.
 * Returns false if another transmission is pending and the new buffering
 * cannot be performed.
 *
 * Strictly, you don't need to call begin(), you can just call write(),
 * but begin makes the caller's code clearer, and allows you to wait
 * until the buffer is ready with:
 *  while(!begin());
 */
bool I2C_Class::begin() {
	if (_masterTxBufferLength != 0) {
		// we are already transmitting (at least one character is pending
		// from a previous non-blocking tx, or because some data has been
		// buffered with write)
		return false;
	}
	return true;
}

/*
 * Sends the bytes buffered by write(). If 'sendStop' == false,
 * does not release the bus, so that the master is guaranteed to be able
 * to talk again.
 *
 * Output   I2C_ERR_NONE
 *          or I2C_ERR_ARB_LOST
 *          or I2C_ERR_ADDR_NACK
 *          or I2C_ERR_DATA_NACK
 *          or I2C_ERR_BUS
 *          or I2C_ERR_UNKNOWN
 *          or ERR_PENDING_TX
 *          or ERR_RESERVED_ADDR
 */
i2c_err I2C_Class::send(uint8_t address, bool sendStop) {
	// addresses 1~7 and 119~127 are reserved; but allow to transmit to them,
	// someone might need it.
	// address 0 indicates a general call
	
    if(!isValidAddress(address, true)) {
	    return ERR_RESERVED_ADDR;
	}

	if (_masterTxBufferIndex > 0) {
		// the buffer is already in use
		return ERR_PENDING_TX;
	}

	// transmit buffer (blocking)
	i2c_err ret = doMasterTx(address, sendStop, true);
	// reset tx buffer iterator vars
	_masterTxBufferIndex = 0;
	_masterTxBufferLength = 0;
	return ret;
}

/*
 * Sends the bytes buffered by write(), and the caller is not
 * blocked here until the transmission is over.
 * This way the caller is free for other processing, but it has no way
 * to know if the message is received.
 * Also, a stop bit is always sent, to avoid locking the bus.
 *
 * Returns ERR_PENDING_TX
 *         or ERR_RESERVED_ADDR
 *         or I2C_ERR_NONE
 *
 * Since the call is non-blocking, I2C_ERR_NONE actually means "we didn't wait
 * for the response, so we don't know if everything went well: let's hope it did".
 *
 * NOTE: this function must still wait until the bus is idle, then it start
 * transmitting; instead of waiting for the transmission to be over, it
 * immediately returns.
 */
i2c_err I2C_Class::sendNonBlocking(uint8_t address) {
	// addresses >119 are reserved; but allow to transmit to them, someone
	// might need it

	if(!isValidAddress(address,true)) {
		return ERR_RESERVED_ADDR;
    }

	if (_masterTxBufferIndex > 0) {
		// the buffer is already in use
		return ERR_PENDING_TX;
	}

	doMasterTx(address, true, false);
	return I2C_ERR_NONE;
}

/*
 * Write a single byte.
 *
 * ERR_PENDING_TX = another masterTx is pending
 *  0 = buffer is already full
 *  1 = ok
 */
int I2C_Class::write(uint8_t data) {
	if (_masterTxBufferIndex > 0) {
		// the buffer is already in use
		return ERR_PENDING_TX;
	}

	if (_masterTxBufferLength >= MAX_MASTER_TX_BUFFER_LENGTH) {
		return 0;
	}

	_masterTxBuffer[_masterTxBufferLength++] = data;
        
	return 0;
}

/*
 * Write multiple bytes (not more that can fit into the buffer).
 *
 * ERR_PENDING_TX = another masterTx is pending
 *  0 = buffer is already full
 *  N = number of bytes that fit into the buffer
 */
int I2C_Class::write(const uint8_t *data, size_t quantity) {
	if (_masterTxBufferIndex > 0) {
		// the buffer is already in use
		return ERR_PENDING_TX;
	}

	if (_masterTxBufferLength >= MAX_MASTER_TX_BUFFER_LENGTH) {
		return 0;
	}

	size_t leftQty = MAX_MASTER_TX_BUFFER_LENGTH - _masterTxBufferLength;
	if (quantity > leftQty) {
		quantity = leftQty;
	}

	memcpy(&_masterTxBuffer[_masterTxBufferLength], data, quantity);
	_masterTxBufferLength += quantity;

	return quantity;
}

/*
 * Write bytes until zero-terminator is reached (or until the buffer is full).
 *
 * ERR_PENDING_TX = another masterTx is pending
 *  0 = buffer is already full
 *  N = number of bytes that fit into the buffer
 *
 *  If data is too long, the terminator will not be sent.
 */
int I2C_Class::write(const char* data, bool sendTerminator) {
	return write(reinterpret_cast<const uint8_t*>(data), strlen(data) + (sendTerminator ? 1 : 0));
}

/*
 * Write a single byte in response to the master's request.
 * This function can be called only inside the onSlaveRequest callback.
 *
 * ERR_NOT_STX = not in slave tx mode (called outside the callback)
 *  0 = buffer is already full
 *  1 = ok
 */
int I2C_Class::writeResponse(uint8_t data) {
	if (I2C_STATE_STX != _state) {
		return ERR_NOT_STX;
	}

	if (_slaveTxBufferLength >= MAX_SLAVE_TX_BUFFER_LENGTH) {
		return 0;
	}

	_slaveTxBuffer[_slaveTxBufferLength++] = data;
	return 1;
}

/*
 * Write multiple bytes (not more that can fit into the buffer)
 * in response to the master's request.
 * This function can be called only inside the onSlaveRequest callback.
 *
 * ERR_NOT_STX = not in slave tx mode (called outside the callback)
 *  0 = buffer is already full
 *  N = number of bytes that fits into the buffer
 */
int I2C_Class::writeResponse(const uint8_t *data, size_t quantity) {
	if (I2C_STATE_STX != _state) {
		return ERR_NOT_STX;
	}

	if (_slaveTxBufferLength >= MAX_SLAVE_TX_BUFFER_LENGTH) {
		return 0;
	}

	size_t leftQty = MAX_SLAVE_TX_BUFFER_LENGTH - _slaveTxBufferLength;
	if (quantity > leftQty) {
		quantity = leftQty;
	}

	memcpy(&_slaveTxBuffer[_slaveTxBufferLength], data, quantity);
	_slaveTxBufferLength += quantity;

	return quantity;
}

/*
 * Write bytes in response to the master's request, until zero-terminator
 * is reached (or until the buffer is full).
 * This function can be called only inside the onSlaveRequest callback.
 *
 * ERR_NOT_STX = not in slave tx mode (called outside the callback)
 *  0 = buffer is already full
 *  N = number of bytes that fits into the buffer
 *
 *  If data is too long, the terminator will not be sent.
 */
int I2C_Class::writeResponse(const char* data, bool sendTerminator) {
	return writeResponse(reinterpret_cast<const uint8_t*>(data),
			strlen(data) + (sendTerminator ? 1 : 0));
}

/*
 * Waits until the bus is idle, then try and become a master to send a request
 * to the selected slave (can't send requests to 0x00).
 *
 * Length is the number of bytes to read. They will be available in _masterRxBuffer.
 * The master always read 'length' bytes. If the slave sends less bytes,
 * garbage is read.
 *
 * Output   >=0 : the number of bytes read
 *          or I2C_ERR_ARB_LOST
 *          or I2C_ERR_ADDR_NACK
 *          or I2C_ERR_BUS
 *          or I2C_ERR_UNKNOWN
 *          or I2C_ERR_INVALID_GCALL
 */
i2c_err I2C_Class::doMasterRequestTx(uint8_t address, uint8_t length, bool sendStop) {
	if (address == 0x00) {
		return I2C_ERR_GCALL_REQUEST;
	}
	
	// ensure data will fit into buffer
	if (MAX_MASTER_RX_BUFFER_LENGTH < length) {
		length = MAX_MASTER_RX_BUFFER_LENGTH;
	}

	// wait until twi is ready, become master receiver
  #if I2C_TIMEOUT_MS == 0
    while (I2C_STATE_READY != _state) {
        continue;
    }
  #else
    if(I2C_STATE_READY != _state) {
        uint32_t t0 = millis();
	    while (I2C_STATE_READY != _state && millis() - t0 < I2C_TIMEOUT_MS) {
		    continue;
	    }
        
        if(I2C_STATE_READY != _state) { // Abort
            resetBus();
            releaseBus();
            _error = TW_NO_INFO;
            return I2C_ERR_TIMEOUT;
        }
    }
  #endif
    
	_state = I2C_STATE_MRX;
	_sendStop = sendStop;
	// reset error state
	_error = TW_NO_INFO;

	// initialize buffer iteration vars
	_masterRxBufferIndex = 0;
	_masterRxBufferLength = length;

	// build sla+w, slave device address + w bit
	_slarw = (address << 1) | TW_READ;

	if (_inRepStart) {
		// if we're in the repeated start state, then we've already sent the start,
		// (@@@ we hope), and the TWI statemachine is just waiting for the address byte.
		// We need to remove ourselves from the repeated start state before we enable interrupts,
		// since the ISR is ASYNC, and we could get confused if we hit the ISR before cleaning
		// up. Also, don't enable the START interrupt. There may be one pending from the
		// repeated start that we sent outselves, and that would really confuse things.
		_inRepStart = false;		// remember, we're dealing with an ASYNC ISR
		TWDR = _slarw;
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN) | _BV(TWIE);// enable INTs, but not START
	} else
		// send start condition
		TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT) | _BV(TWSTA);

	// wait for read operation to complete
  #if I2C_RESET_TIMEOUT_MS == 0
      while (I2C_STATE_MRX == _state) {
          continue;
      }
  #else
    if(I2C_STATE_MRX == _state) {
        uint32_t t0 = millis();
	    while (I2C_STATE_MRX == _state && millis() - t0 < I2C_TIMEOUT_MS) {
		    continue;
	    }
        
        if(I2C_STATE_MRX == _state) { // Abort
            resetBus();
            releaseBus();
            _error = TW_NO_INFO;
            return I2C_ERR_TIMEOUT;
        }
    }
  #endif

	if (_masterRxBufferIndex < length) {
		length = _masterRxBufferIndex;
    }

	switch (_error) {
	case TW_NO_INFO: // success
		return length;
	case TW_MT_ARB_LOST: // MR_ARB_LOST handled as MT_ARB_LOST
		return I2C_ERR_ARB_LOST;
	case TW_MR_SLA_NACK:
		return I2C_ERR_ADDR_NACK;
	case TW_BUS_ERROR:
		return I2C_ERR_BUS;
	default:
		return I2C_ERR_UNKNOWN;
	}
}

/*
 * Waits until the bus is idle, then try and become a master to send data to
 * a slave (or as general call, if address is 0x00).
 *
 * NOTE: the function always waits for the bus to be idle, then starts
 * transmitting. The 'wait' parameter only dictates if the function must wait
 * for the bus to become idle again, or to return immediately.
 * Even if 'wait' is true, the first waiting is be done (before acquiring the channel).
 *
 * The data to be transmitted must already be inside _masterTxBuffer.
 *
 * Output   I2C_ERR_NONE
 *          or I2C_ERR_ARB_LOST
 *          or I2C_ERR_ADDR_NACK
 *          or I2C_ERR_DATA_NACK
 *          or I2C_ERR_BUS
 *          or I2C_ERR_UNKNOWN
 */
i2c_err I2C_Class::doMasterTx(uint8_t address, bool sendStop, bool wait) {
    // wait until twi is ready, become master transmitter
  #if I2C_TIMEOUT_MS == 0
    while (I2C_STATE_READY != _state) {
        continue;
    }
  #else
	if(I2C_STATE_READY != _state) {
    	uint32_t t0 = millis();
    	while (I2C_STATE_READY != _state && millis() - t0 < I2C_TIMEOUT_MS) {
        	continue;
    	}
    	
    	if(I2C_STATE_READY != _state) { // Abort
        	resetBus();
            releaseBus();
        	_error = TW_NO_INFO;
        	_masterTxBufferLength = 0;
        	_masterTxBufferIndex = 0;
        	return I2C_ERR_TIMEOUT;
    	}
	}
  #endif
	_state = I2C_STATE_MTX;
	_sendStop = sendStop;
	// reset error state
	_error = TW_NO_INFO;

	// build sla+w, slave device address + w bit
	_slarw = (address << 1) | TW_WRITE;

	// if we're in a repeated start, then we've already sent the START
	// in the ISR. Don't do it again.
	//
	if (_inRepStart) {
		// if we're in the repeated start state, then we've already sent the start,
		// (@@@ we hope), and the TWI statemachine is just waiting for the address byte.
		// We need to remove ourselves from the repeated start state before we enable interrupts,
		// since the ISR is ASYNC, and we could get confused if we hit the ISR before cleaning
		// up. Also, don't enable the START interrupt. There may be one pending from the
		// repeated start that we sent outselves, and that would really confuse things.
		_inRepStart = false;		// remember, we're dealing with an ASYNC ISR
		TWDR = _slarw;
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN) | _BV(TWIE);// enable INTs, but not START
	} else {
		// send start condition
		TWCR = _BV(TWINT) | _BV(TWEA) | _BV(TWEN) | _BV(TWIE) | _BV(TWSTA);	// enable INTs
	}

	// wait for write operation to complete
  #if I2C_TIMEOUT_MS == 0
    if(wait) {
        while (I2C_STATE_MTX == _state) {
            continue;
        }
    }
  #else
    if(wait) {
	    if(I2C_STATE_MTX == _state) {
    	    uint32_t t0 = millis();
    	    while (I2C_STATE_MTX == _state && millis() - t0 < I2C_TIMEOUT_MS) {
        	    continue;
    	    }
    	
    	    if(I2C_STATE_MTX == _state) { // Abort
        	    resetBus();
                releaseBus();
        	    _error = TW_NO_INFO;
        	    _masterTxBufferLength = 0;
        	    _masterTxBufferIndex = 0;
        	    return I2C_ERR_TIMEOUT;
    	    }
	    }
    }
  #endif

	switch (_error) {
	case TW_NO_INFO: // success
		return 0;
	case TW_MT_ARB_LOST:
		return I2C_ERR_ARB_LOST;
	case TW_MT_SLA_NACK:
		return I2C_ERR_ADDR_NACK;
	case TW_MT_DATA_NACK:
		return I2C_ERR_DATA_NACK;
	case TW_BUS_ERROR:
		return I2C_ERR_BUS;
	default:
		return I2C_ERR_UNKNOWN;
	}
}

void I2C_Class::sendWithAck() const {
	// transmit master read ready signal, with ack
	// data has already been loaded into TWDR
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWINT) | _BV(TWEA);
}

void I2C_Class::sendWithNack() const {
	// transmit master read ready signal, without ack
	// data has already been loaded into TWDR
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWINT);
}

/*
 * Send a stop bit to end the transmission and release the bus.
 */
i2c_err I2C_Class::sendStop(void) {
	// send stop condition
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT) | _BV(TWSTO);

	// wait for stop condition to be executed on bus
	// TWINT is not set after a stop condition!
  #if I2C_STOP_TIMEOUT_US == 0
    while (TWCR & _BV(TWSTO)) {
        continue;
    }
  #else
    if(TWCR & _BV(TWSTO)) {
        uint32_t t0 = micros();
        while (TWCR & _BV(TWSTO) && micros() - t0 < I2C_STOP_TIMEOUT_US) {
            continue;
        }
        
        if(TWCR & _BV(TWSTO)) { // Abort
            resetBus();
            releaseBus();
            _error = TW_NO_INFO;
            return I2C_ERR_TIMEOUT;
        }
    }
  #endif
    
	// update twi state
	_state = I2C_STATE_READY;
    
    return I2C_ERR_NONE;
}

/*
 * Release the bus without sending a stop bit.
 */
void I2C_Class::releaseBus(void) {
	// release bus
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA) | _BV(TWINT);

	// update twi state
	_state = I2C_STATE_READY;
}

void I2C_Class::resetBus() {
    // use a freq of 10kHz, that should never be too fast; it may be slow, but the reset
    // is not a common operation, so we can bear it.
    const uint32_t delayUS = 100;
    
    pinMode(SDA, _pullups ? INPUT_PULLUP : INPUT);
    
    uint32_t t0 = millis();
    while((I2C_RESET_TIMEOUT_MS == 0 || millis() - t0 < I2C_RESET_TIMEOUT_MS) && LOW == digitalRead(SDA)) {
        pinMode(SCL, OUTPUT);
        // clock pulse
        digitalWrite(SCL, HIGH); delayMicroseconds(delayUS);
        digitalWrite(SCL, LOW);  delayMicroseconds(delayUS);
        
        if(HIGH == digitalRead(SDA)){
            // send stop bit manually
            pinMode(SDA, OUTPUT);
            digitalWrite(SDA, LOW);   delayMicroseconds(delayUS); // hold SDA low, needed for stop condition
            digitalWrite(SCL, HIGH);  delayMicroseconds(delayUS); // stop condition
            digitalWrite(SDA, HIGH);  delayMicroseconds(delayUS); // also release SDA
            pinMode(SDA, _pullups ? INPUT_PULLUP : INPUT);
        } else {
            // complete clock pulse
            digitalWrite(SCL, HIGH); delayMicroseconds(delayUS);
        }
    }
    
    pinMode(SDA, _pullups ? INPUT_PULLUP : INPUT);
    pinMode(SCL, _pullups ? INPUT_PULLUP : INPUT);
    
    releaseBus(); // to reset the TWI state
}

/*
 * Handles all TWI events: address rx/tx, data rx/tx, ack/nack handling, etc.
 */
void I2C_Class::onInterrupt() {
	switch (TW_STATUS) {
	// All Master
	case TW_START:     // sent start condition
	case TW_REP_START: // sent repeated start condition
		// copy device address and r/w bit to output register and ack
		TWDR = _slarw;
		sendWithAck();
		break;

		// Master Transmitter
	case TW_MT_SLA_ACK:  // slave receiver acked address
	case TW_MT_DATA_ACK: // slave receiver acked data
		// if there is data to send, send it, otherwise stop
		if (_masterTxBufferIndex < _masterTxBufferLength) {
			// copy data to output register and ack
			TWDR = _masterTxBuffer[_masterTxBufferIndex++];
			sendWithAck();
		} else {
			if (_sendStop) {
				if(sendStop() != I2C_ERR_NONE) {
                    _error = TW_BUS_ERROR;
                }
			} else {
				_inRepStart = true;	// we're gonna send the START
				// don't enable the interrupt. We'll generate the start, but we
				// avoid handling the interrupt until we're in the next transaction,
				// at the point where we would normally issue the start.
				TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);
				_state = I2C_STATE_READY;
			}
			_masterTxBufferLength = 0;
			_masterTxBufferIndex = 0;
		}
		break;
	case TW_MT_SLA_NACK:  // address sent, nack received
		_error = TW_MT_SLA_NACK;
		_masterTxBufferLength = 0;
		_masterTxBufferIndex = 0;
		if(sendStop() != I2C_ERR_NONE) {
    		_error = TW_BUS_ERROR;
		}
		break;
	case TW_MT_DATA_NACK: // data sent, nack received
		_error = TW_MT_DATA_NACK;
		_masterTxBufferLength = 0;
		_masterTxBufferIndex = 0;
		if(sendStop() != I2C_ERR_NONE) {
    		_error = TW_BUS_ERROR;
		}
		break;
	case TW_MT_ARB_LOST: // lost bus arbitration
		//or case TW_MR_ARB_LOST:
		_error = TW_MT_ARB_LOST;
		_masterTxBufferLength = 0;
		_masterTxBufferIndex = 0;
		releaseBus();
		break;

		// Master Receiver
	case TW_MR_DATA_ACK: // data received, ack sent
		// put byte into buffer
	    _masterRxBuffer[_masterRxBufferIndex++] = TWDR;
		/* no break */
	case TW_MR_SLA_ACK:  // address sent, ack received
		// ack if more bytes are expected, otherwise nack
		if (_masterRxBufferIndex < _masterRxBufferLength - 1) {
			sendWithAck();
		} else {
			sendWithNack();
		}
		break;
	case TW_MR_DATA_NACK: // data received, nack sent
		// put final byte into buffer
  		_masterRxBuffer[_masterRxBufferIndex++] = TWDR;
		_masterRxBufferLength = 0;
		//_masterRxBufferIndex = 0;
		// ^^ this is still needed to know the actual number of bytes received
		// it must be cleared by the caller (doMasterRequestTx)!!
		if (_sendStop) {
			if(sendStop() != I2C_ERR_NONE) {
    			_error = TW_BUS_ERROR;
			}
		} else {
			_inRepStart = true;	// we're gonna send the START
			// don't enable the interrupt. We'll generate the start, but we
			// avoid handling the interrupt until we're in the next transaction,
			// at the point where we would normally issue the start.
			TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);
			_state = I2C_STATE_READY;
		}
		break;
	case TW_MR_SLA_NACK: // address sent, nack received
		_error = TW_MR_SLA_NACK;
		if(sendStop() != I2C_ERR_NONE) {
    		_error = TW_BUS_ERROR;
		}
		_masterRxBufferLength = 0;
		_masterRxBufferIndex = 0;
		break;

    // case TW_MR_ARB_LOST: handled by TW_MT_ARB_LOST case

		// Slave Receiver
	case TW_SR_SLA_ACK:   // addressed, returned ack
	case TW_SR_GCALL_ACK: // addressed generally, returned ack
	case TW_SR_ARB_LOST_SLA_ACK:   // lost arbitration, returned ack
	case TW_SR_ARB_LOST_GCALL_ACK: // lost arbitration, returned ack

		if (TW_STATUS == TW_SR_GCALL_ACK || TW_STATUS == TW_SR_ARB_LOST_GCALL_ACK) {
			_gCall = true;
		} else {
			_gCall = false;
		}
		// enter slave receiver mode
		_state = I2C_STATE_SRX;
		// indicate that rx buffer can be overwritten and ack
		_slaveRxBufferIndex = 0;
		sendWithAck();
		break;
	case TW_SR_DATA_ACK:       // data received, returned ack
	case TW_SR_GCALL_DATA_ACK: // data received generally, returned ack
		// if there is still room in the rx buffer
		if (_slaveRxBufferIndex < MAX_SLAVE_RX_BUFFER_LENGTH) {
			// put byte in buffer and ack
			_slaveRxBuffer[_slaveRxBufferIndex++] = TWDR;
			sendWithAck();
		} else {
			// otherwise nack
			sendWithNack();
		}
		break;
	case TW_SR_STOP: // stop or repeated start condition received
		// put a null char after data if there's room
		if (_slaveRxBufferIndex < MAX_SLAVE_RX_BUFFER_LENGTH) {
			_slaveRxBuffer[_slaveRxBufferIndex] = '\0';
		}
		// sends ack and stops interface for clock stretching
		if(sendStop() != I2C_ERR_NONE) {
    		_error = TW_BUS_ERROR;
		}

		if (_error == TW_NO_INFO && _slaveOnReceive) {
			// callback to user defined callback, only if it has been attached
			// pass the buffer and its length
			_slaveOnReceive(_slaveRxBuffer, _slaveRxBufferIndex, _gCall);
		}
		// reset the buffer for future data
		_slaveRxBufferIndex = 0;
		// ack future responses and leave slave receiver state
		releaseBus();
		break;
	case TW_SR_DATA_NACK:       // data received, returned nack
	case TW_SR_GCALL_DATA_NACK: // data received generally, returned nack
		// nack back at master
		sendWithNack();
		break;

		// Slave Transmitter
	case TW_ST_SLA_ACK:          // addressed, returned ack
	case TW_ST_ARB_LOST_SLA_ACK: // arbitration lost, returned ack
		// enter slave transmitter mode
		_state = I2C_STATE_STX;
		// ready the tx buffer index for iteration
		_slaveTxBufferIndex = 0;
		// set tx buffer length to be zero, to verify if user changes it
		_slaveTxBufferLength = 0;
		// request for txBuffer to be filled and length to be set
		if (_slaveOnRequest) {
			_slaveOnRequest();
		}
		// transmit first byte from buffer, fallthrough intended
		/* no break */
	case TW_ST_DATA_ACK: // byte sent, ack returned
		// copy data to output register
		TWDR = (_slaveTxBufferIndex < MAX_SLAVE_TX_BUFFER_LENGTH)
               ? _slaveTxBuffer[_slaveTxBufferIndex++]
               : 0x00;
		// if there is more to send, ack, otherwise nack
		if (_slaveTxBufferIndex < _slaveTxBufferLength) {
			sendWithAck();
		} else {
			sendWithNack();
		}
		break;
	case TW_ST_DATA_NACK: // received nack, we are done
	case TW_ST_LAST_DATA: // received ack, but we are done already!
		// ack future responses
		sendWithAck();
		// leave slave receiver state
		_state = I2C_STATE_READY;
		break;

		// All
	case TW_NO_INFO:   // no state information
		break;
	case TW_BUS_ERROR: // bus error, illegal stop/start
		_error = TW_BUS_ERROR;
		sendStop();
		break;
	}
}

I2C_Class I2C;

SIGNAL(TWI_vect) {
	I2C.onInterrupt();
}
