/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This library has been adapted from the 'Wire' (and 'twi') library from
 * Arduino's "official" libraries.
 *
 * Improvements over 'Wire':
 * + Buffer size can be selected (for each of the four SRX, STX, MRX, MTX modes)
 * + Frequency can be selected
 * + Can enable/disable internal pullups
 * + Handles general calls
 * + Can enable/disable the channel (the device will be unreachable, but
 *   will still be able to send data to other devices)
 * + Reduced memory footprint (data won't be copied multiple times: the
 *   tx message is still composed with begin-write-send as before, but is not copied
 *   through multiple library levels; the rx message is stored in one single
 *   location, directly accessible by the caller, that will copy it only if needed)
 * + Improved speed: the library makes less copies of the data and when it makes copy
 *   it does it faster
 *   (e.g. when transmitting messages of 32 bytes @400kHz, the cumulative time
 *   [channel_transmission + processing_by_the_library] is about 20% smaller;
 *   in other scenarios the gain is not so high, since the channel transmission time
 *   depends little on the library)
 */

/* The original Wire/twi license disclaimer follows:
 ===================================================
 TwoWire.h - TWI/I2C library for Arduino & Wiring
 [twi.h - TWI/I2C library for Wiring & Arduino]
 Copyright (c) 2006 Nicholas Zambetti.  All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 Modified 2012 by Todd Krein (todd@krein.org) to implement repeated starts
 */

/*
 * Class to manage two-wire (or I2C) communication.
 *
 * Master Tx:
 * + Call begin(). If it returns false, another tx is pending, so
 *   you should wait (write())
 * + Call write() as many times as needed to buffer the message
 * + Call send() to send the buffered message. This can be blocking (the caller waits
 *   for the transmission to complete) or non-blocking.
 *   If non-blocking, calling begin() when this transmission is still going on will
 *   return false (cannot buffer multiple transmissions)
 * + To send data to all devices enabled for "general call", send the message
 *   to the address 0x00 (which can't be the valid address of a device)
 *
 * Slave Rx:
 * + Register a callback to attachSlaveOnReceive() to be notified
 * + The callback will receive the data (array of bytes, with its size)
 *
 * Slave Tx:
 * + Register a callback to attachSlaveOnRequest() to be notified when
 *  the master requests data.
 * + Call writeResponse() as many times as needed (can do that ONLY inside the callback)
 * + When the callback returns, the data will be sent
 * + NOTE: the master expects at least one byte to be returned. If the callback doesn't
 *   write anything, automatically a 0x00 is sent. (this can be used to indicate that
 *   0 bytes were actually to be sent; see the next note)
 * + NOTE: even if the slave doesn't transmit enough bytes, the master has no way to
 *   know it, so it will read garbage (probably 0xFF, because of the idle bus).
 *   The application should provide some other way to know the length, e.g.:
 *   - The master transmit a special packet (as masterTx) "Request of N bytes coming soon..."
 *   - The master transmit a request packet (as masterRequestTx); the transmission
 *     library must know the expected number of bytes (the exceeding bytes won't be
 *     asked to the slave, so they will never be transmitted)
 *   - The slave answer with a packet where the first byte indicates the number
 *     of bytes: this way the master is able to know if it has asked too many bytes
 *     and can discard the final garbage bytes
 *
 * Master Rx:
 * + Call requestFrom()
 * + It will return the slave's response
 *
 * Stop bits:
 * + To release the bus, the master sends a stop bit; if he doesn't, he still
 *   holds the channel and is guaranteed that no one else can interfere (no
 *   arbitration).
 *
 * Usage:
 * + First, set all parameters needed
 *   - Call defaultSettings() to set all settings. This is needed if you don't
 *   want to specify the frequency or the buffer lengths, since they cannot be left
 *   unspecified
 *   - Call enablePullups() or disablePullups()
 *     NOTE: if you disable pullups and don't connect external pullup resistors,
 *     the bus will probably fall to a low level, effectively blocking every
 *     transmission.
 *   - Call setAddress() to set this device's address (the address must be in 0x08~0x77 (8~119);
 *    0x00 is reserved for general call; 0x00~0x07 and 0x78~0x7F (0~7 and 120~127) are reserved)
 *    To be able to use these addresses too, you must enable it with I2C_USE_RESERVED_ADDRESSES.
 *     If you don't want to receive data, set the address to 0x00. If you want to
 *     receive only general calls, set the address to 0x00 and enable general calls.
 *   - Call receiveGeneralCalls() if you want
 *   - Call setBufferLength() to set slaveRx/slaveTx/masterRx/masterTx buffers
 *   - Register slave callbacks on receive data/on receive request, to
 *    attachSlaveOnReceive() and attachSlaveOnRequest()
 *    NOTE: data can be safely read from the channel only inside the callback made by
 *    attachSlaveOnReceive()
 * + Call enable() to enable listening to the channel. If you send data, the channel will
 *   be automatically enabled (it will be enabled during send(), not before); after
 *   that you have to disable it manually.
 *   When the channel is disabled, no one can send data to this device (the master will receive
 *   NACK); instead, if it is enabled but no callback is set, the data will be discarded,
 *   but the master will receive an ACK.
 *   To enable general calls only, or to enable tx only, just set the address to 0x00:
 *   no one will be able to send data to this device (except for general calls, if
 *   enabled).
 * + Send/receive data ...
 * + Call disable() to disable the channel; you can then re-enable it with enable().
 *   NOTE: you don't have to disable the channel to complete the transmission
 * + If the bus seems to freeze, you can call releaseBus(). This may not solve the
 *   problem, if other devices are locking it, and may also interrupt valid transmission,
 *   possibly causing unedefined behaviors.
 *   The best thing to do to avoid bus freezes is to use the callbacks to do as
 *   little work as possible: set some flags, copy the useful data, and perform
 *   heavy calculations in another loop (NOT just another function called by
 *   the callback, but a function called by the main loop).
 */

#ifndef I2C_H_
#define I2C_H_

#include <Arduino.h>
#include <inttypes.h>
#include <stddef.h>
#include <stdlib.h>

#include "bitWrite.h"

typedef int16_t i2c_err;

#ifndef I2C_TIMEOUT_MS
    // The time after which a communication is considered stuck (locked slave, plug/unplug problems).
    // The communication is aborted and a resetBus() is issued
    #define I2C_TIMEOUT_MS 50 // 0 = no timeout
#endif

#ifndef I2C_RESET_TIMEOUT_MS
    // The time after which resetBus() is aborted (it wasn't successful; maybe someone is shorting a line
    // to ground, or there is another non-recoverable problem).
    #define I2C_RESET_TIMEOUT_MS 100 // 0 = no timeout
#endif

#ifndef I2C_STOP_TIMEOUT_US
    // The time after which sendStop() is aborted.
    #define I2C_STOP_TIMEOUT_US 50000 // 0 = no timeout
#endif

#define I2C_USE_RESERVED_ADDRESSES true

// Possible states for the channel
#define I2C_STATE_READY 0
#define I2C_STATE_MRX   1
#define I2C_STATE_MTX   2
#define I2C_STATE_SRX   3
#define I2C_STATE_STX   4

// Possible errors (must be negative to distinguish them from the I2C states)
#define I2C_ERR_NONE           ((i2c_err)   0) // no error (any >=0 value is a no-error value; sometimes values >=0 have a specific meaning, e.g. a number of bytes)
#define I2C_ERR_ADDR_NACK      ((i2c_err)  -1) // nack received on address sent (maybe the device on that address doesn't exist)
#define I2C_ERR_DATA_NACK      ((i2c_err)  -2) // nack received on data
#define I2C_ERR_ARB_LOST       ((i2c_err)  -3) // tried to send data, but another master transmitted simultaneously and took the control of the bus
#define I2C_ERR_BUS            ((i2c_err)  -4) // invalid start/stop sequence, ...
#define I2C_ERR_UNKNOWN        ((i2c_err) -10)
#define I2C_ERR_GCALL_REQUEST  ((i2c_err) -11) // trying to send a request to general call: impossible, slaves can't answer together!
#define I2C_ERR_TIMEOUT        ((i2c_err) -12) // to signal a timeout in the communication
#define ERR_PENDING_TX         ((i2c_err) -20) // trying to tx when another tx by this same device is still pending (the library cannot queue transmissions)
#define ERR_NOT_STX            ((i2c_err) -21) // trying to send a response to a master request when outside the onRequest callback
#define ERR_RESERVED_ADDR      ((i2c_err) -30) // trying to use an invalid address (if this check is enabled by I2C_USE_RESERVED_ADDRESSES)

class I2C_Class {
public:
	// You must NEVER use this to create new instances!
	// Use the already allocated instance 'I2C' instead.
	I2C_Class();
	~I2C_Class();

    static bool isValidAddress(uint8_t addr, bool gcallOk = true); // if gcallOk == true, addr == 0 will return true; for standard addresses the result depends on I2C_USE_RESERVED_ADDRESSES too
	void defaultSettings();
	bool setFreq(uint32_t frequency = 100000L);
	bool setBufferLength(uint8_t buffLen = 32);
	bool setBufferLength(uint8_t slaveRxLen, uint8_t slaveTxLen,
			uint8_t masterRxLen, uint8_t masterTxLen);
	bool setAddress(uint8_t address);
	void receiveGeneralCalls(bool enable = true);
	void enablePullups();
	void disablePullups();
	void attachSlaveOnReceive(
			void (*callback)(const uint8_t* buff, uint8_t num_bytes,
					bool generalCall));
	void attachSlaveOnRequest(void (*callback)());

	void enable();
	void disable();
	bool begin();
	i2c_err send(uint8_t address, bool sendStop = true);
	i2c_err sendNonBlocking(uint8_t address);
	int write(uint8_t data);
	int write(const uint8_t* data, size_t quantity);
	int write(const char* data, bool sendTerminator = false);

	int writeResponse(uint8_t data);
	int writeResponse(const uint8_t* data, size_t quantity);
	int writeResponse(const char* data, bool sendTerminator = false);

	i2c_err requestFrom(uint8_t address, uint8_t*& data, uint8_t quantity,
			bool sendStop = true);

    /* Just release the bus without sending a stop bit (the bus may hang because of this) */
	void releaseBus();
    /*
     * Release the bus by sending clock pulses until the SDA is lifted and then send a stop bit
     * (this should unlock the bus if some slave is waiting for a master transmission that will never end)
     */
    void resetBus();

	// you must NEVER call this directly, it is called during I2C interrupts
	inline void onInterrupt();
    
private:
    bool _pullups; // internal pullups enabled/disabled
	volatile uint8_t _state;
	volatile uint8_t _slarw; // 7bit address + TW_READ or TW_WRITE
	volatile uint8_t _sendStop; // should the transaction end with a stop
	volatile uint8_t _inRepStart; // in the middle of a repeated start
	volatile uint8_t _error;
	volatile bool _gCall; // true if the last slaveRx was a general call

	// user callbacks:
	// + when the slave has received data from the master
	// + when the slave must reply to the master's request
	// They are volatile pointers to function, so that the callbacks can be
	// changed by the user at any moment.
	void (* volatile _slaveOnReceive)(const uint8_t* buff, uint8_t num_bytes,
			bool generalCall);
	void (* volatile _slaveOnRequest)();

	uint8_t MAX_SLAVE_RX_BUFFER_LENGTH;
	uint8_t MAX_SLAVE_TX_BUFFER_LENGTH;
	uint8_t MAX_MASTER_RX_BUFFER_LENGTH;
	uint8_t MAX_MASTER_TX_BUFFER_LENGTH;

	uint8_t* _slaveRxBuffer;
	volatile uint8_t _slaveRxBufferIndex;
	// length not needed: as new bytes are received, index is incremented: index==length

	uint8_t* _slaveTxBuffer;
	volatile uint8_t _slaveTxBufferIndex; // incremented as the buffered data is sent over the channel
	volatile uint8_t _slaveTxBufferLength; // incremented as the caller buffers data

	uint8_t* _masterRxBuffer;
	volatile uint8_t _masterRxBufferIndex; // incremented as new data arrive: finally, it's index==length
	volatile uint8_t _masterRxBufferLength; // EXPECTED length (as specified by master's request)

	uint8_t* _masterTxBuffer;
	volatile uint8_t _masterTxBufferIndex; // incremented as the buffered data is sent over the channel
	volatile uint8_t _masterTxBufferLength; // incremented as the caller buffers data

	I2C_Class(const I2C_Class&); // unused
	I2C_Class& operator=(const I2C_Class&); // unused

	i2c_err doMasterRequestTx(uint8_t address, uint8_t length, bool sendStop =
			true);
	i2c_err doMasterTx(uint8_t address, bool sendStop = true, bool wait = true);
	void sendWithAck() const;
	void sendWithNack() const;
	i2c_err sendStop();
};

////////////////////////////////////////////////////////////////

inline
bool I2C_Class::isValidAddress(uint8_t addr, bool gcallOk) {
    if(addr == 0) {
        return gcallOk;
    }
    
#if I2C_USE_RESERVED_ADDRESSES
    return addr < 128;
#else
    return addr >= 8 && addr <= 119;
#endif
}

inline
I2C_Class::~I2C_Class() {
	free(_slaveRxBuffer);
	free(_slaveTxBuffer);
	free(_masterRxBuffer);
	free(_masterTxBuffer);
}

/*
 * Sets the function that is called as soon as a complete packet
 * (data between a start and a stop bit) is received.
 * Data can be safely read only inside this callback.
 *
 * If 0 is passed as function, the data will still be read from the channel,
 * but discarded.
 *
 * You can call this function at any moment.
 */
inline
void I2C_Class::attachSlaveOnReceive(
		void (*callback)(const uint8_t* buff, uint8_t num_bytes,
				bool generalCall)) {
	_slaveOnReceive = callback;
}

/*
 * Sets the function that is called as soon as a request is received.
 * The response must be written inside this callback; it is sent as soon as
 * the callback returns.
 *
 * If 0 is passed as function, or if no response is written by the callback,
 * a single 0x00 byte is sent to the master (plus garbage bytes).
 *
 * You can call this function at any moment.
 */
inline
void I2C_Class::attachSlaveOnRequest(void (*callback)()) {
	_slaveOnRequest = callback;
}

/*
 * Set the length for the four buffers.
 *
 * Returns false if there's not enough memory.
 * NOTE: if the memory is not enough, all buffers are deallocated.
 * Do not call this while the channel is enabled or some tx message bytes
 * have already been buffered.
 */
inline
bool I2C_Class::setBufferLength(uint8_t buffLen) {
	return setBufferLength(buffLen, buffLen, buffLen, buffLen);
}

inline
void I2C_Class::enablePullups() {
    _pullups = true;
	digitalWrite(SDA, HIGH);
	digitalWrite(SCL, HIGH);
}

inline
void I2C_Class::disablePullups() {
    _pullups = false;
	digitalWrite(SDA, LOW);
	digitalWrite(SCL, LOW);
}

inline
void I2C_Class::receiveGeneralCalls(bool enable) {
	BSET(enable, TWAR, _BV(TWGCE));
}
/*
 * Enable the channel (enable receiving data; transmission is automatically
 * enabled as it is started).
 */
inline
void I2C_Class::enable() {
	// enable twi module, acks, and twi interrupt
	TWCR = _BV(TWEN) | _BV(TWIE) | _BV(TWEA);
}

/*
 * Disable the channel (disable receiving data; transmission is automatically
 * enabled as it is started).
 */
inline
void I2C_Class::disable() {
	TWCR = 0x00;
}

extern I2C_Class I2C;

#endif /* I2C_H_ */
