/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#ifndef IMGCHUNKER_H_
#define IMGCHUNKER_H_

#include "fmatdef.h"
#include <inttypes.h>
#include <string.h>

namespace FlashMat {
/*
 * Used to compose a bigger image out of 8x8 sub-images or viceversa, handling the proper
 * color depth.
 * This class does not perform color depth conversions.
 *
 * 'col' and 'row' are the coordinates of the 8x8 subimage.
 * If the row/col is outside the boundaries of the image, nothing is drawn (for set)
 * and a black image is returned (for get).
 */

class ImgChunker {
public:
    static void getChunk(uint8_t dst[/*SIZE_8x8_FOR(D)*/], const uint8_t src[/*FRAME_SIZE_FOR(D)*/], uint8_t col, uint8_t row, uint8_t colorDepth);
    static void setChunk(uint8_t dst[/*FRAME_SIZE_FOR(D)*/], const uint8_t src[/*SIZE_8x8_FOR(D)*/], uint8_t col, uint8_t row, uint8_t colorDepth);

    static void getChunk8bit(uint8_t dst[SIZE_8x8_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(8)], uint8_t col, uint8_t row);
    static void getChunk4bit(uint8_t dst[SIZE_8x8_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(4)], uint8_t col, uint8_t row);
    static void getChunk2bit(uint8_t dst[SIZE_8x8_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(2)], uint8_t col, uint8_t row);
    static void getChunk1bit(uint8_t dst[SIZE_8x8_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(1)], uint8_t col, uint8_t row);
    
    static void setChunk8bit(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[SIZE_8x8_FOR(8)], uint8_t col, uint8_t row);
    static void setChunk4bit(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[SIZE_8x8_FOR(4)], uint8_t col, uint8_t row);
    static void setChunk2bit(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[SIZE_8x8_FOR(2)], uint8_t col, uint8_t row);
    static void setChunk1bit(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[SIZE_8x8_FOR(1)], uint8_t col, uint8_t row);
    
private:
    ImgChunker(); // unused
};
///////////////////////////////////////

inline
void ImgChunker::getChunk8bit(uint8_t dst[SIZE_8x8_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(8)], uint8_t col, uint8_t row) {
    getChunk(dst, src, col, row, 8);
}

inline
void ImgChunker::getChunk4bit(uint8_t dst[SIZE_8x8_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(4)], uint8_t col, uint8_t row) {
    getChunk(dst, src, col, row, 4);
}

inline
void ImgChunker::getChunk2bit(uint8_t dst[SIZE_8x8_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(2)], uint8_t col, uint8_t row) {
    getChunk(dst, src, col, row, 2);
}

inline
void ImgChunker::getChunk1bit(uint8_t dst[SIZE_8x8_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(1)], uint8_t col, uint8_t row) {
    getChunk(dst, src, col, row, 1);
}

inline
void ImgChunker::setChunk8bit(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[SIZE_8x8_FOR(8)], uint8_t col, uint8_t row) {
    setChunk(dst, src, col, row, 8);
}

inline
void ImgChunker::setChunk4bit(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[SIZE_8x8_FOR(4)], uint8_t col, uint8_t row) {
    setChunk(dst, src, col, row, 4);
}

inline
void ImgChunker::setChunk2bit(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[SIZE_8x8_FOR(2)], uint8_t col, uint8_t row) {
    setChunk(dst, src, col, row, 2);
}

inline
void ImgChunker::setChunk1bit(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[SIZE_8x8_FOR(1)], uint8_t col, uint8_t row) {
    setChunk(dst, src, col, row, 1);
}

inline
void ImgChunker::getChunk(uint8_t dst[/*SIZE_8x8_FOR(D)*/],
                          const uint8_t src[/*FRAME_SIZE_FOR(D)*/],
                          uint8_t col, uint8_t row,
                          uint8_t colorDepth)
{
    const uint8_t x0 = 8 * col;
    const uint8_t y0 = 8 * row;
    
    const uint8_t rowSize1col = colorDepth; // 8 pixels  *  D bit/pixel  /  8 bit/byte
    
    if(x0 >= MATRIX_COLS || y0 >= MATRIX_ROWS) {
        // outside the boundaries of this module's frame
        memset(dst, 0x00, SIZE_8x8_FOR(colorDepth)); // fill with black
        return;
    }
    
    // iterate over color planes
    // iterate over rows (each 8-pixel row has contiguous pixels;
    //   different rows are not contiguous unless the whole matrix is an 8x8)
    // for each row, copy it bit-by-bit
    for(uint8_t row = 0; row < 8; row++) {
        for (uint8_t c = 0; c < 3; c++) {
            const uint16_t srcIndex = (x0 + MATRIX_COLS * ((y0 + row) + MATRIX_ROWS * (uint16_t)c)) * colorDepth / 8;
            const uint16_t dstIndex = (               8 * (      row  +           8 * (uint16_t)c)) * colorDepth / 8;
            
            memcpy(&dst[dstIndex], &src[srcIndex], rowSize1col);
        }
    }
}

inline
void ImgChunker::setChunk(uint8_t dst[/*FRAME_SIZE_FOR(D)*/],
                          const uint8_t src[/*SIZE_8x8_FOR(D)*/],
                          uint8_t col, uint8_t row,
                          uint8_t colorDepth)
{
    const uint8_t x0 = 8 * col;
    const uint8_t y0 = 8 * row;
    
    const uint8_t rowSize1col = colorDepth; // 8 pixels  *  D bit/pixel  /  8 bit/byte
    
    if(x0 >= MATRIX_COLS || y0 >= MATRIX_ROWS) {
        // outside the boundaries of this module's frame
        return;
    }
    
    // iterate over color planes
    // iterate over rows (each 8-pixel row has contiguous pixels;
    //   different rows are not contiguous unless the whole matrix is an 8x8)
    // for each row, copy it bit-by-bit
    for(uint8_t row = 0; row < 8; row++) {
        for (uint8_t c = 0; c < 3; c++) {
            const uint16_t dstIndex = (x0 + MATRIX_COLS * ((y0 + row) + MATRIX_ROWS * (uint16_t)c)) * colorDepth / 8;
            const uint16_t srcIndex = (               8 * (      row  +           8 * (uint16_t)c)) * colorDepth / 8;
            
            memcpy(&dst[dstIndex], &src[srcIndex], rowSize1col);
        }
    }
}

} // namespace

#endif /* IMGCHUNKER_H_ */