/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#ifndef IMG_DEPTH_CONVERTER_H_
#define IMG_DEPTH_CONVERTER_H_

#include "fmatdef.h"
#include <inttypes.h>

namespace FlashMat {
/*
 * All color depth conversions are performed front-to-back, so that in case of conversions from a smaller
 * to a higher color depth, the source can be stored in the last bytes of the destination, without
 * risk of overwriting yet-to-be-used data.
 * Note that src is const, but it is overwritten if you use this buffering technique.
 * Instead, if you store the src in another memory location, it will be left unchanged.
 */

class ImgDepthConverter {
public:
    static void c8to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(8)]);
    static void c8to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(8)]);
    static void c8to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(8)]);
    static void c4to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(4)]);
    static void c4to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(4)]);
    static void c4to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(4)]);
    static void c2to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(2)]);
    static void c2to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(2)]);
    static void c2to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(2)]);
    static void c1to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(1)]);
    static void c1to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(1)]);
    static void c1to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(1)]);

private:
    ImgDepthConverter(); // unused
};
///////////////////////////////////////

inline
void ImgDepthConverter::c8to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(8)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(4);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        *(dst++) = ((data0 & 0xF0) >> 0)
                 | ((data1 & 0xF0) >> 4); // pack 2 values into 1 byte, using their MSBs
    }
}

inline
void ImgDepthConverter::c8to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(8)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(2);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        uint8_t data2 = *(src++);
        uint8_t data3 = *(src++);
        *(dst++) = ((data0 & 0xC0) >> 0)
                 | ((data1 & 0xC0) >> 2)
                 | ((data2 & 0xC0) >> 4)
                 | ((data3 & 0xC0) >> 6); // pack 4 values into 1 byte, using their MSBs
    }
}

inline
void ImgDepthConverter::c8to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(8)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(1);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        uint8_t data2 = *(src++);
        uint8_t data3 = *(src++);
        uint8_t data4 = *(src++);
        uint8_t data5 = *(src++);
        uint8_t data6 = *(src++);
        uint8_t data7 = *(src++);
        *(dst++) = ((data0 & 0x80) >> 0)
                 | ((data1 & 0x80) >> 1)
                 | ((data2 & 0x80) >> 2)
                 | ((data3 & 0x80) >> 3)
                 | ((data4 & 0x80) >> 4)
                 | ((data5 & 0x80) >> 5)
                 | ((data6 & 0x80) >> 6)
                 | ((data7 &0x80) >> 7); // pack 8 values into 1 byte, using their MSBs
    }
}

inline
void ImgDepthConverter::c4to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(4)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(4);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 2 values of 4 bits each
        *(dst++) = (data & 0xF0) | (data >> 4); // first value expanded 4->8 (replicated on LSBs)
        *(dst++) = (data << 4) | (data & 0x0F); // second value expanded 4->8 (replicated on LSBs)
    }
}

inline
void ImgDepthConverter::c4to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(4)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(2);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        *(dst++) = ((data0 & 0xC0) >> 0)
                 | ((data0 & 0x0C) << 2)
                 | ((data1 & 0xC0) >> 4)  // pack 4 values into 1 byte, using their MSBs
                 | ((data1 & 0x0C) >> 2); // (they are already packed into 2 bytes)
    }
}

inline
void ImgDepthConverter::c4to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(4)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(1);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        uint8_t data2 = *(src++);
        uint8_t data3 = *(src++);
        *(dst++) = ((data0 & 0x80) >> 0)
                 | ((data0 & 0x08) << 3)
                 | ((data1 & 0x80) >> 2)
                 | ((data1 & 0x08) << 1)
                 | ((data2 & 0x80) >> 4)
                 | ((data2 & 0x08) >> 1)
                 | ((data3 & 0x80) >> 6)  // pack 8 values into 1 byte, using their MSBs
                 | ((data3 & 0x08) >> 3); // (they are already packed into 4 bytes)
    }
}

inline
void ImgDepthConverter::c2to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(2)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(2);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 4 values of 2 bits each
        *(dst++) = ((data & 0xC0) >> 0) | ((data & 0xC0) >> 2)
                 | ((data & 0xC0) >> 4) | ((data & 0xC0) >> 6); // first value expanded 2->8 (replicated on LSBs)
        *(dst++) = ((data & 0x30) << 2) | ((data & 0x30) >> 0)
                 | ((data & 0x30) >> 2) | ((data & 0x30) >> 4); // second value expanded 2->8 (replicated on LSBs)
        *(dst++) = ((data & 0x0C) << 4) | ((data & 0x0C) << 2)
                 | ((data & 0x0C) >> 0) | ((data & 0x0C) >> 2); // third value expanded 2->8 (replicated on LSBs)
        *(dst++) = ((data & 0x03) << 6) | ((data & 0x03) << 4)
                 | ((data & 0x03) << 2) | ((data & 0x03) >> 0); // fourth value expanded 2->8 (replicated on LSBs)
    }
}

inline
void ImgDepthConverter::c2to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(2)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(2);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 4 values of 2 bits each
        *(dst++) = ((data & 0xC0) >> 0) | ((data & 0xC0) >> 2)  // first value expanded 2->4 (replicated on LSBs)
                 | ((data & 0x30) >> 2) | ((data & 0x30) >> 4); // second value expanded 2->4 (replicated on LSBs)
        *(dst++) = ((data & 0x0C) << 4) | ((data & 0x0C) << 2)  // third value expanded 2->4 (replicated on LSBs)
                 | ((data & 0x03) << 2) | ((data & 0x03) >> 0); // fourth value expanded 2->4 (replicated on LSBs)
    }
}

inline
void ImgDepthConverter::c2to1(uint8_t dst[FRAME_SIZE_FOR(1)], const uint8_t src[FRAME_SIZE_FOR(2)]){
    const uint8_t* dstEnd = dst + FRAME_SIZE_FOR(1);
    for (; dst < dstEnd;) {
        uint8_t data0 = *(src++);
        uint8_t data1 = *(src++);
        *(dst++) = ((data0 & 0x80) >> 0)
                 | ((data0 & 0x20) << 1)
                 | ((data0 & 0x08) << 2)
                 | ((data0 & 0x02) << 3)
                 | ((data1 & 0x80) >> 4)
                 | ((data1 & 0x20) >> 3)
                 | ((data1 & 0x08) >> 2)  // pack 8 values into 1 byte, using their MSBs
                 | ((data1 & 0x02) >> 1); // (they are already packed into 2 bytes)
    }
}

inline
void ImgDepthConverter::c1to8(uint8_t dst[FRAME_SIZE_FOR(8)], const uint8_t src[FRAME_SIZE_FOR(1)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(1);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 8 values of 1 bit each
        *(dst++) = ((data & 0x80) ? 0xFF : 0x00); // first value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x40) ? 0xFF : 0x00); // second value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x20) ? 0xFF : 0x00); // third value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x10) ? 0xFF : 0x00); // fourth value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x04) ? 0xFF : 0x00); // fifth value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x03) ? 0xFF : 0x00); // sixth value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x02) ? 0xFF : 0x00); // seventh value expanded 1->8 (replicated on LSBs)
        *(dst++) = ((data & 0x01) ? 0xFF : 0x00); // eighth value expanded 1->8 (replicated on LSBs)
    }
}

inline
void ImgDepthConverter::c1to4(uint8_t dst[FRAME_SIZE_FOR(4)], const uint8_t src[FRAME_SIZE_FOR(1)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(1);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 8 values of 1 bit each
        *(dst++) = ((data & 0x80) ? 0xF0 : 0x00)  // first value expanded 1->4 (replicated on LSBs)
                 | ((data & 0x40) ? 0x0F : 0x00); // second value expanded 1->4 (replicated on LSBs)
        *(dst++) = ((data & 0x20) ? 0xF0 : 0x00)  // third value expanded 1->4 (replicated on LSBs)
                 | ((data & 0x10) ? 0x0F : 0x00); // fourth value expanded 1->4 (replicated on LSBs)
        *(dst++) = ((data & 0x08) ? 0xF0 : 0x00)  // fifth value expanded 1->4 (replicated on LSBs)
                 | ((data & 0x04) ? 0x0F : 0x00); // sixth value expanded 1->4 (replicated on LSBs)
        *(dst++) = ((data & 0x02) ? 0xF0 : 0x00)  // seventh value expanded 1->4 (replicated on LSBs)
                 | ((data & 0x01) ? 0x0F : 0x00); // eigth value expanded 1->4 (replicated on LSBs)
    }
}

inline
void ImgDepthConverter::c1to2(uint8_t dst[FRAME_SIZE_FOR(2)], const uint8_t src[FRAME_SIZE_FOR(1)]){
    const uint8_t* srcEnd = src + FRAME_SIZE_FOR(1);
    for (; src < srcEnd; src++) {
        uint8_t data = *src; // 1 byte is made of 8 values of 1 bit each
        *(dst++) = ((data & 0x80) >> 0) | ((data & 0x80) >> 1)  // first value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x40) >> 1) | ((data & 0x40) >> 2)  // second value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x20) >> 2) | ((data & 0x20) >> 3)  // third value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x10) >> 3) | ((data & 0x10) >> 4); // fourth value expanded 1->2 (replicated on LSBs)
        *(dst++) = ((data & 0x08) << 4) | ((data & 0x08) << 3)  // fifth value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x04) << 3) | ((data & 0x04) << 2)  // sixth value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x02) << 2) | ((data & 0x02) << 1)  // seventh value expanded 1->2 (replicated on LSBs)
                 | ((data & 0x01) << 1) | ((data & 0x01) << 0); // eighth value expanded 1->2 (replicated on LSBs)
    }
}

} // namespace
#endif /* IMG_DEPTH_CONVERTER_H_ */
