/*
 * Copyright 2013 C&D Elettronica. All rights reserved.
 *
 * This file is part of the FlashMat project.
 */

#ifndef TEXT_H_
#define TEXT_H_

#include <inttypes.h>
#include <string.h>

#include "fmatdef.h"
#include "Point.h"
#include "Font.h"
#include "Fonts.h"
#include "safe_string.h"

namespace FlashMat {

class Text {
public:
	// The string of text to be shown
	char str[TEXT_BUFFER_SIZE + 1];
	// The foreground color
	uint8_t color[3];
	// The background color
	uint8_t bgColor[3];
	// The origin of the entire string
	coord_t topleft[2];
	// The char/line spacing
	uint8_t charSpacing;
	uint8_t lineSpacing;

	// Spacing type for char/line
	// false = adaptive = use character-specific width / font-specific height
	// true = monospace = each char is in its 8x8 cell, as defined by the font
	bool monospace;

	// True if the text is to be drawn over the current image
	// False if the text is to be drawn over a uniform bgColor background
	bool overlay;

    Fonts::FontId fontId; // just a nice place to hold the id for the user (the cell doesn't need it), to send it over through the commander
	const Font* font;

	Text();
            
    /*
     * Calculates the space occupation of the text with all the
     * current settings (of course, 'topleft' doesn't matter).
     *
     * If you don't need an argument, provide a null argument.
     *
     * Width and height are always >=0, but having them signed
     * makes the caller's life easier when calculating things like
     * "available_space - actual_width", where signed arithmetics is needed
     */
    void getStringPixelSize(coord_t* width, coord_t* height) const;
    Point getStringPixelSize() const;
    
    void setText(const char str[]);
    void setText(const uint8_t str[], uint8_t len);
    void setText(const __FlashStringHelper* str);

    void setColor(const uint8_t color[3]);
    void setBgColor(const uint8_t color[3]);

    void setTopleft(const coord_t topleft[2]);
    void setTopleft(coord_t x, coord_t y);
    
    void setCharSpacing(uint8_t space);
    void setLineSpacing(uint8_t space);
    void setMonospace(bool mono = true);
    void setOverlay(bool ovlay = true);
    
    void setFont(Fonts::FontId id);
};


inline
void Text::setText(const char str[]) {
    strcpy_s(this->str, str, TEXT_BUFFER_SIZE);
}

inline
void Text::setText(const __FlashStringHelper* str) {
    for(uint16_t i = 0; i < TEXT_BUFFER_SIZE; i++) {
        this->str[i] = pgm_read_byte(reinterpret_cast<const uint8_t*>(str) + i);
        
        if(this->str[i] == '\0') return;
    }
}

inline
void Text::setText(const uint8_t str[], uint8_t len) {
  #if TEXT_BUFFER_SIZE < 255
    if(len > TEXT_BUFFER_SIZE) {
        len = TEXT_BUFFER_SIZE;
    }
  #endif
    
    memcpy(this->str, str, len);
    this->str[len] = '\0';
}

inline
void Text::setColor(const uint8_t color[3]) {
    memcpy(this->color, color, 3);
}

inline
void Text::setBgColor(const uint8_t color[3]) {
    memcpy(this->bgColor, color, 3);
}

inline
void Text::setTopleft(const coord_t topleft[2]) {
    setTopleft(topleft[X], topleft[Y]);
}

inline
void Text::setTopleft(coord_t x, coord_t y) {
    topleft[X] = x;
    topleft[Y] = y;
}

inline
void Text::setCharSpacing(uint8_t space) {
    this->charSpacing = space;
}

inline
void Text::setLineSpacing(uint8_t space) {
    this->lineSpacing = space;
}

inline
void Text::setMonospace(bool mono) {
    this->monospace = mono;
}

inline
void Text::setOverlay(bool ovlay) {
    this->overlay = ovlay;
}

inline
void Text::setFont(Fonts::FontId id) {
    this->fontId = id;
    this->font = Fonts::getFont(id);
}

}
#endif /* TEXT_H_ */
