/*
 Copyright (C) 2013-2014 C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Example usage of the FlashMat commander to send drawing commands to FlashMat cells.
 * You must upload this code to an Arduino, open a serial communication from your PC to the
 * Arduino and send it the number corresponding to the example test you want to perform.
 * See the "switch(test)" inside loop() to know which test numbers are available.
 *
 * This is a possible order in which you can perform the tests:
 * + Choose the 'address' variable below to be one of your cell. If you don't know it,
 *   just ignore it, you'll discover it later (and then you'll update the code and upload it
 *   again to the Arduino).
 * + Connect the commander to the cells (these pins are needed: GND, SDA, SCL).
 *   Cells can be cascaded using their strip connectors (directly, to connect them side-by-side;
 *   if you want to tile them vertically, just connect their right-side connectors using
 *   wires).
 * + Run testPing() to check that the cell is responsive and has the right address.
 * + If you have unknown cells, or something is not working, run testDiscover() to
 *   know which addresses are set for the connected cells. You cannot tell which address
 *   corresponds to which cell, unless you connect just one cell at a time.
 * + If you need to reconfigure the address of a cell, connect only that cell to the commander,
 *   and run testCellAddress(). Then run testDiscover() again to check that everything's all right.
 * + Run the tests meant for the first cell only (the first ones listed).
 * + Run testMultiCell() (inside the test it is specified which addresses and
 *   which configuration is required).
 * + Run the advanced tests testStartup(), testWait(), testSwapTypes(), testTiming()
 */

#include "FlashMat_master.h"

using namespace FlashMat;

// NOTE: if you create any function that has a FlashMat's class (Point, Gradient, etc.) as an argument
// or as the return value, you MUST declare it here.
// e.g. if you have:
//     Text getMyDefaultTextParameters(int myPar1, bool myPar2) {
//         Text myTextPars;
//         ... some stuff to set the parameters ...
//         return myTextPars;
//     }
// you must add here this declaration:
//     Text getMyDefaultTextParameters(int myPar1, bool myPar2);
// The declaration must match EXACTLY the definition (same types, same names for the parameters, same
// spaces), and they must be placed here, before any function and after the 'using namespace' line.
// If you forget to do that, you'll get obscure errors from Arduino IDE, like "'Text' does not name a type"
// on a line that doesn't even have "Text" in it (the problem is caused by some transformations that
// the Arduino does on your .ino file automatically before verifying/compiling it, and sometimes
// it gets them wrong).

// Set the cell address you are using.
// Tests that use more than one cell will define their own addresses (check their function definitions).
uint8_t address = 64; // for maximum compatibility, this must be in [8,119]

// DECLARATIONS
////////////////////////////////////////////////////////////////////////////////

void printHelp();

void prompt();
void prompt2();
void promptAnyKeyToStop();

bool waitAnyKey();
void cleanSerial();

void printStatus(std::pair<CellStatus, PacketStatus> status);

void testPing();
void testDiscover();
void testCellAddress();

void testFill();
void testDrawPixel();
void testDrawLineH();
void testDrawLineV();
void testDrawRect();
void testDrawGradient();
void testDrawRainbow();

void testDrawText();
void testBitmap();

void testMultiCell();

void testWait();
void testStartup();
void testSwapTypes();
void testTiming();

void testIdentify();

void blankAll();

//////////////////////////////////

void drawLemming(uint8_t img[3][MATRIX_ROWS][MATRIX_COLS], int frame, bool overlay = false);

#define printName(str) Serial.println(F(str))

// DEFINITIONS
////////////////////////////////////////////////////////////////////////////////

const uint8_t W = 255;

const uint8_t black  [3] = MAKE_RGB(0,0,0);
const uint8_t white  [3] = MAKE_RGB(W,W,W);
const uint8_t red    [3] = MAKE_RGB(W,0,0);
const uint8_t green  [3] = MAKE_RGB(0,W,0);
const uint8_t blue   [3] = MAKE_RGB(0,0,W);
const uint8_t yellow [3] = MAKE_RGB(W,W,0);
const uint8_t purple [3] = MAKE_RGB(W,0,W);
const uint8_t cyan   [3] = MAKE_RGB(0,W,W);
const uint8_t orange [3] = MAKE_RGB(W,W/2,0);
const uint8_t violet [3] = MAKE_RGB(W/2,0,W);
const uint8_t lime   [3] = MAKE_RGB(W/2,W,0);
const uint8_t pink   [3] = MAKE_RGB(W,W/2,W/2);

////////////////////////////////////////

void setup() {
    // To receive the number of the test to be performed, and to return info to the user
    Serial.begin(9600);
    
    Serial.print(F(
        "FlashMat master example." "\n"
        "Initializing (and waiting for the I2C bus to become available)..." "\n"
    ));
    
    // Init the hardware needed to send drawing commands to the FlashMat cells
    commander.setup();
    
    // Configure the object that will send drawing commands
    // For the tests, this configuration is ok. To know more about the configuration (and about
    // the commander's behavior), look at testStartup() and testWait().
    commander.selectCell(address);  // The address to which we send commands (we can change it whenever we need)
    commander.setTimeoutUs(5000);   // The time after which a packet is dropped if the cell is not there to receive it
    commander.setPacketDelayUs(0);  // The time to wait after each packet, to avoid sending packets too often (so that the cell has enough spare time to actually do its job)
    commander.setNumRetry(1);       // Attempts (plus the initial one) made to send a packet; if all of them fail, the packet is dropped
    commander.setWaitTimeoutMs(200); // Max time (ms) for which to wait for a cell to become idle again
    
    delay(1000); // Give the cells some time to power up and set up, before we send any command to them
    
    printHelp();
    
    I2C.setFreq(400000L);
}

void printHelp() {
    Serial.println();
    Serial.print(F(
        "Choose a test number: " "\n"
        "  1. ping" "\n"
        "  2. discover" "\n"
        "  3. change the address of the connected cells / change the address used for the other tests" "\n"
        " 10. fill" "\n"
        " 11. draw pixel" "\n"
        " 12. draw line H" "\n"
        " 13. draw line V" "\n"
        " 14. draw rect" "\n"
        " 15. draw gradient" "\n"
        " 16. draw rainbow" "\n"
        " 17. draw text" "\n"
        " 18. draw bitmap" "\n"
        " 30. test multicell" "\n"
        " 50. wait" "\n"
        " 51. startup" "\n"
        " 52. swap types" "\n"
        " 53. timing" "\n"
        " 90. identify" "\n"
        "100. blank all" "\n"
        "  H. print this help again" "\n"
    ));

    Serial.println();
}

/*
 * Here we receive a number over the serial (we interpret it from its ascii representation)
 * and then we perform the corresponding test.
 * Whenever we are waiting for a number, the user will see a '>' prompt on the serial.
 */
void loop() {
    static uint8_t test = 0; // the test number (0 means we are waiting for a number)
    bool parsed = false; // true when we received a new number
    
    // At startup, prompt the user
    static bool first = true;
    if(first) {
        first = false;
        prompt();
    }
    
    // Check if a number is available
    if(Serial.available()) {
        char c = Serial.peek();
        if(c == 'h' || c == 'H') {
            printHelp();
            prompt();
        } else {
            delay(50); // wait some time, so that all ASCII characters have been received (otherwise we would truncate the number)
            long i = parseLong(Serial); // interpret the number from the received characters
            
            // Check the number correctness
            if(i <= 0 || i > 255) {
                Serial.print(F("Test number must be in [1-255] (given: "));
                Serial.print(i);
                Serial.println(')');
                prompt();
            } else {
                test = i;
                parsed = true;
            }
        }
        
        cleanSerial(); // discard any remaining characters (clean up the serial buffer)
    }
    
    // Feedback to the user, to confirm that we received the number
    if(parsed) {
        Serial.print(F("Test #"));
        Serial.print(test);
        Serial.print(F(": "));
    }
    
    // Actually perform the test
    switch(test) {
        case   1: testPing();             break;
        case   2: testDiscover();         break;
        case   3: testCellAddress();      break;
        
        case  10: testFill();             break;
        case  11: testDrawPixel();        break;
        case  12: testDrawLineH();        break;
        case  13: testDrawLineV();        break;
        case  14: testDrawRect();         break;
        case  15: testDrawGradient();     break;
        case  16: testDrawRainbow();      break;
        case  17: testDrawText();         break;
        case  18: testBitmap();           break;

        case  30: testMultiCell();        break;

        case  50: testWait();             break;
        case  51: testStartup();          break;
        case  52: testSwapTypes();        break;
        case  53: testTiming();           break;
        
        case  90: testIdentify();         break;
        case 100: blankAll();             break; // just to turn off all the matrices if they become annoyingly bright
        
        default:
            if(parsed) {
                Serial.println(F("No available test for this number"));
            }
            break;
    }
    
    // If we just completed a test, we get back into the waiting state and prompt the user again
    if(test != 0) {
        test = 0;
        cleanSerial(); // remove characters given while a test was running
        prompt();
    }
}

void prompt() {
    Serial.println('>');
}

void prompt2() {
    Serial.println(F(">>>"));
}

void promptAnyKeyToStop() {
    Serial.println(F("Any key to stop"));
}

bool waitAnyKey() {
    cleanSerial();               // empty the serial buffer (remove old input, if any)
    while(!Serial.available());  // wait for at least one new character to arrive
    delayMs(20);                 // wait for all the characters to arrive (or most of them, at least)
    char c = Serial.read();      // read the first character
    cleanSerial();               // empty the serial buffer
    return c == 'q' || c == 'Q'; // we can use 'Q' to quit a test
}

void cleanSerial() {
    while(Serial.read() >= 0);
}

void printStatus(std::pair<CellStatus, PacketStatus> status) {
    switch(status.first) {
        case IDLE:
            Serial.print(F("IDLE"));
            break;
        case BUSY:
            Serial.print(F("BUSY"));
            break;
        default:
            //Serial.print(F("UNKNOWN"));
            Serial.print('<');
            Serial.print(int(status.first));
            Serial.print('>');
            break;
    }
    
    Serial.print(' ');
    
    switch(status.second) {
        case ACCEPTED:
            Serial.print(F("ACCEPTED"));
            break;
        case REJECTED:
            Serial.print(F("REJECTED"));
            break;
        case PENDING:
            Serial.print(F("PENDING"));
            break;
        default:
            //Serial.print(F("UNKNOWN"));
            Serial.print('<');
            Serial.print(int(status.second));
            Serial.print('>');
            break;
    }
}

/*
 * Shows how to set the address for a cell, even if the cell hasn't got an address yet.
 */
void testCellAddress() {
    printName("setCellAddress");
    
    /*
     * To set the address for a cell, we send it with the proper command.
     * In testMultiCell() it is shown that to send the same command to many cells
     * you must select their addresses with Commander::setBcastAddresses().
     * But here the cells may have no address at all; then sendStoreAddress()
     * will send the packet to ANY cell that is connected (there is an I2C address, 0,
     * to do that).
     */
    
    Serial.print(F(
        "Enter a number for an address [8-119]." "\n"
        "This will be set as the address for any connected cell" "\n"
        "  (the cell itself will change its internally-stored address)." "\n"
        "Note that any following test will be performed on this new address." "\n"
        "If you just want to change the address for tests, use this command with no cells connected" "\n"
        "Enter 'Q' to quit" "\n"
    ));
    
    for(;;) {
        prompt2();
        while(!Serial.available()); // wait for the user input

        char c = Serial.peek();
    
        if(c == 'q' || c == 'Q') {
            // exit the test
            break;
        } else {
            // it must be an address
            delay(20);
            bool parsed = false;
            long val = parseLong(Serial, &parsed);
            if(!parsed || (val < 8 || val > 119)) {
                Serial.print(F("The address must be in [8,119] (given: "));
                Serial.print(val);
                Serial.println(')');
            } else {
                // address ok: change it and exit the test
                address = val;
                Serial.print(F("Changing address to "));
                Serial.println(address);
                commander.selectCell(BROADCAST); // update all the connected cells
                commander.sendStoreAddress(address, 0, 0); // only the first argument is used for now
                commander.selectCell(address);
                break;
            }
        }
    
        cleanSerial();
    }
    
    cleanSerial(); // empty the serial buffer
}

/*
 * Check if the cell is there, active and responsive to the configured address.
 */
void testPing() {
    printName("ping");

    // Tell the commander which cell we need
    commander.selectCell(address);

    // Check if the cell exists and responds to this address
    int16_t res = commander.ping();
    
    // Whenever we send a packet, we get an error in response.
    // Anything different from I2C_ERR_NONE indicates failure.
    Serial.print(F("Ping #"));
    Serial.print(int(address));
    Serial.print(F(": "));
    Serial.println(res == I2C_ERR_NONE ? F("ok") : F("no response"));
}

/*
 * Scan the I2C bus to find which addresses are active.
 */
void testDiscover() {
    printName("discover");

    Serial.println(F("Found:"));
    
    int16_t res;
    
    uint8_t total = 0; // number of active cells
    bool bcast = false; // did anyone respond to broadcast?
    
    // Check if anyone is listening to broadcast
    commander.selectCell(BROADCAST);
    res = commander.ping();
    if(res == I2C_ERR_NONE) {
        bcast = true;
        Serial.print('\t');
        Serial.print((int)BROADCAST);
        Serial.println(F(" (BROADCAST)"));
    }
    
    // Check for any valid address
    for(uint8_t i = 8; i <= 119; ++i) {
        commander.selectCell(i);
        res = commander.ping();
        
        if(res == I2C_ERR_NONE){
            ++total;
            Serial.print('\t');
            Serial.println((int)i);
        }
    }
    
    Serial.print(F("\tTotal: "));
    Serial.print((int)total);
    
    if(bcast) {
        Serial.print(F(" (plus someone on broadcast)"));
    }
    
    Serial.println();
    
    // Restore the old address
    commander.selectCell(address);
}

/*
 * Fill the matrix in one color and then in another, after a fixed delay.
 * Test for the packets:
 * + PKT_FILL
 * + PKT_SWAP
 */
void testFill(){
    printName("fill");
    
    commander.selectCell(address);
    
    /*
     * The FlashMat cell uses a double-buffer strategy.
     * The current image is stored in the 'front buffer'; the firmware that is
     * responsible for sending the color data to the hardware, to show it, takes
     * its data from the front buffer.
     * Meanwhile, the software that is drawing the new image (from the received
     * drawing commands) draws it in the 'back buffer'. When the new image is complete,
     * a 'swap' drawing command causes the two buffers to swap: now the matrix-driving
     * software will take its data from the new front buffer (which used to be the back
     * buffer), so that it shows the new image, and the drawing software can process
     * new drawing commands by drawing on the new back buffer.
     * This way, the shown image is still, i.e. the watcher doesn't see the image
     * forming piece by piece as it is being drawn: he sees the old image until the new
     * one is ready, and then it suddenly sees the new one.
     * Without double-buffering, the watcher would see ugly-looking 'artifacts', that
     * can be as bad as causing the sequence of images to appear almost as noise or as
     * malfunctioning.
     *
     * In summary, to draw an image we need to:
     * + Draw it on the back buffer, using whatever commands we like (we can chain as many
     *   commands as we like; the new ones will draw over the current partially-drawn image)
     * + Show it by sending a 'swap' command. The back buffer will be swapped with the front
     *   buffer: the just-completed image will be shown, and the former front buffer will be
     *   now the back buffer, ready to be drawn to.
     *   (To control what image is in this new back buffer just after the swap, before any
     *   command is received, see testSwapTypes() )
     */
    
    commander.sendFill(yellow);
    commander.sendSwap();
    
    delay(500);
    
    commander.sendFill(cyan);
    commander.sendSwap();
    
    delay(500);
}

/*
 * Draws one pixel at a time.
 * Test for the packets:
 * + PKT_DRAW_PIXEL
 * + PKT_COPY_BUFFER
 * + PKT_SWAP
 */
void testDrawPixel() {
    printName("drawPixel");

    promptAnyKeyToStop();
    
    const uint8_t delayMs = 40;
    
    cleanSerial();
    
    commander.selectCell(address);
    
    // Draw one pixel at a time
    for(int i = 0; i < MATRIX_COLS; i++) {
        if(Serial.available()) {
            return;
        }
        
        delay(delayMs);
        commander.sendFill(black); // Blank out the matrix
        commander.sendDrawPixel(i, 0, red); // draw the pixel
        commander.sendSwap(); // show the new image
    }
    
    // The same as above, but vertically
    for(int i = 0; i < MATRIX_ROWS; i++){
        if(Serial.available()) {
            return;
        }
        
        delay(delayMs);
        commander.sendFill(black);
        commander.sendDrawPixel(0, i, green);
        commander.sendSwap();
    }
    
    // Fill with pixels, whole matrix
    // If we just draw the new pixel, we get strange behavior (because we are not writing over
    // the shown image); to actually write over the current image to "add" a pixel at a time,
    // we must first copy the shown image, then add the pixel and then show it.
    commander.sendFill(black);
    for(int y = 0; y < MATRIX_ROWS; y++) {
        for(int x = 0; x < MATRIX_COLS; x++) {
            if(Serial.available()) {
                return;
            }
            
            delay(delayMs);
            commander.sendDrawPixel(x, y, blue); // add a pixel to the current back-buffer image
            commander.sendSwap(); // show the new image
            commander.sendCopyBuffer(); // copy the image for the next iteration
        }
    }
}

/*
 * Draws one row at a time.
 * Test for the packets:
 * + PKT_DRAW_LINE_H
 * + PKT_FILL
 * + PKT_SWAP
 */
void testDrawLineH(){
    printName("drawLineH");

    promptAnyKeyToStop();
    
    commander.selectCell(address);

    for(int i = 0; i < MATRIX_ROWS; i++){
        if(Serial.available()) {
            return;
        }
        
        // Draw one white line over black background
        commander.sendFill(black);
        commander.sendDrawLineH(0, MATRIX_COLS-1, i, white);
        commander.sendSwap();
        
        delay(150);
    }
}

/*
 * Draws one row at a time.
 * Test for the packets:
 * + PKT_DRAW_LINE_V
 * + PKT_FILL
 * + PKT_SWAP
 */
void testDrawLineV(){
    printName("drawLineV");

    promptAnyKeyToStop();
    
    commander.selectCell(address);
    
    for(int i = 0; i < MATRIX_COLS; i++) {
        if(Serial.available()) {
            return;
        }
        
        // Draw purple line over lime background
        commander.sendFill(lime);
        commander.sendDrawLineV(i, 0, MATRIX_ROWS-1, purple);
        commander.sendSwap();
        
        delay(150);
    }
}

/*
 * Draw rectangles.
 * Test for the packets:
 * + PKT_DRAW_RECT
 * + PKT_FILL
 * + PKT_SWAP
 */
void testDrawRect() {
    printName("drawRect");

    promptAnyKeyToStop();
    
    coord_t N = min(MATRIX_ROWS-1, MATRIX_COLS-1);
    
    commander.selectCell(address);
    
    // Rectangles reducing in size towards the center (outline)
    for(coord_t i = 0; i <= N/2; i++) {
        if(Serial.available()) {
            return;
        }
        
        commander.sendFill(black);
        commander.sendDrawRect(i, i, MATRIX_COLS-1-i, MATRIX_ROWS-1-i, yellow, false);
        commander.sendSwap();
        delay(200);
    }
    
    // Rectangles extending horizontally (left to right), filled
    for(coord_t i = 1; i < MATRIX_COLS-1; i++) {
        if(Serial.available()) {
            return;
        }
        
        commander.sendFill(black);
        commander.sendDrawRect(1, 1, i, MATRIX_ROWS-2, orange, true);
        commander.sendSwap();
        delay(150);
    }
}

/*
 * Draw gradients.
 * Test for the packets:
 * + PKT_DRAW_GRADIENT
 * + PKT_SWAP
 */
void testDrawGradient(){
    printName("drawGradient");

    promptAnyKeyToStop();

    // With this array we define the sequence of gradients we want to draw
    const uint8_t N = 14;
    Gradient grads[N] = {GRAD_WEST,
                         GRAD_NORTH_WEST,
                         GRAD_NORTH,
                         GRAD_NORTH_EAST,
                         GRAD_EAST,
                         GRAD_SOUTH_EAST,
                         GRAD_SOUTH,
                         GRAD_SOUTH_WEST,
                         GRAD_DIAMOND,
                         GRAD_CIRCLE,
                         GRAD_SQUARE,
                         GRAD_SQUARE_REV,
                         GRAD_CIRCLE_REV,
                         GRAD_DIAMOND_REV};
    
    Point p0(0,0);
    Point p1(MATRIX_COLS-1, MATRIX_ROWS-1);
    
    commander.selectCell(address);
    
    for(uint8_t i = 0; i < N; i++) {
        if(Serial.available()) {
            return;
        }
        
        commander.sendDrawGradient(yellow, blue, grads[i], p0, p1);
        commander.sendSwap();
        delay(500);
    }
}

/*
 * Draw the 'rainbow' (HSV interpolation with varying H).
 * Test for the packets:
 * + PKT_DRAW_RAINBOW
 * + PKT_SWAP
 */
void testDrawRainbow() {
    /*
     * HSV is a color space, like RGB is.
     * H (hue) defines the "color".
     * S (saturation) defines how vibrant it is ("colored" vs "gray").
     * V (value) defines how bright it is (bright vs black).
     *
     * H varies in [0,360] degrees (or in [0,1], or in [0,255]).
     *       0 = red  ->  60 = yellow -> 120 = green
     *  -> 180 = cyan -> 240 = blue   -> 300 = purple -> 360 = red
     * S varies in [0,1] (or [0,255])
     *    0 = gray (or black/white, depending on V)
     *    1 = vibrant ("very colored")
     * V varies in [0,1] (or [0,255])
     *    0 = black
     *    1 = bright (pure white through pure vibrant, depending on S)
     *
     * =================================================================
     *
     * The rainbow effect is obtained by having a varying H along the
     * horizontal axis, at S=1.
     * Along the vertical axis there is no change in color.
     * You can specify the starting and ending hue, and the effect will
     * interpolate linearly between them (remember that H is periodical,
     * so you can wrap around the red color; also, the effect allows H
     * outside [0,360], to be able to "shift" the rainbow and to have
     * multiple rainbows)
     */
    
    Serial.println(F("Any key to start each test"));
    
    Point p0(0,0);
    Point p1(MATRIX_COLS - 1, MATRIX_ROWS - 1);
    
    commander.selectCell(address);
    
    // Test 1: still rainbow, from red to red (whole range), maximum brightness
    Serial.println(F("Still rainbow 1..."));
    if(waitAnyKey()) { return; }
    commander.sendDrawRainbow(0.0, 360.0, 1.0, p0, p1);
    commander.sendSwap();
    
    // Test 2: still rainbow, from yellow to purple, half brightness
    Serial.println(F("Still rainbow 2..."));
    if(waitAnyKey()) { return; }
    commander.sendDrawRainbow(60.0, 300.0, 0.5, p0, p1);
    commander.sendSwap();
    
    // Test 3: still rainbow, multiple range, going backward, maximum brightness
    Serial.println(F("Still rainbow 3..."));
    if(waitAnyKey()) { return; }
    commander.sendDrawRainbow(720.0, 0.0, 1.0, p0, p1);
    commander.sendSwap();
    
    // Test 4: "moving" rainbow
    Serial.println(F("Moving rainbow..."));
    if(waitAnyKey()) { return; }
    
    // To simulate the movement we simply shift the hue values
    int steps = 60;
    for(int i = 0; i < steps; ++i) {
        float shift = (i / (float)steps) * 360.0;
        // shift = -shift; // to reverse the direction of "movement"
        commander.sendDrawRainbow(0.0 + shift, 360.0 + shift, 1.0, p0, p1);
        commander.sendSwap();
        delayMs(50);
    }
}

/*
 * Draw a text, setting various parameters (message, font, color, ...).
 * Test for the packets:
 * + PKT_FILL
 * + PKT_TEXT
 * + PKT_TEXT_PARS
 * + PKT_TEXT_POSITION
 * + PKT_DRAW_TEXT
 * + PKT_SWAP
 */
void testDrawText() {
    printName("drawText");
    
    // A Text object contains a message and some parameters to display it.
    // We set them later, differently for each test case.
    Text txt;
    
    commander.selectCell(address);
    
    Serial.println(F("Any key to start each test"));
    
    txt.setText(F("Hello"));            // Set the text (max TEXT_BUFFER_SIZE characters)
                                        //     (you can also use '\n' as newlines; any other special character
                                        //      except the terminating '\0' will be displayed as a symbol,
                                        //      or not shown if not supported; none of them has special meaning
                                        //      for the cell)
    txt.setFont(Fonts::FM_8x8);         // The default font (any character fits into 8x8 pixels)
    txt.setColor(red);                  // Color of the letters
    txt.setOverlay(false);              // The text has a uniform background
    txt.setBgColor(blue);               // The background color
    txt.setMonospace(false);            // monospace = each character takes up the same space (in pixels);
                                        //     non-monospace = each character takes up only the needed space (e.g. an 'i' is narrower than a 'w')
    txt.setCharSpacing(1);              // pixels of spacing between characters
    txt.setLineSpacing(1);              // pixels of spacing between text lines
    txt.setTopleft(0,0);                // the starting pixel for the message. Note: if a character is small (it does not occupy the whole
                                        //     vertical space), it may appear that the text is below the desired point; but it is ok, since
                                        //     we must account for tall characters.
    
    // Test 1
    // To draw a text, we must:
    // + Send the text's message
    // + Send the text parameters (color, font, ...)
    // + Send the text position (the topleft xy coordinates; remember that the coordinates start
    //     at [0,0] and they increase towards right and down).
    // + Send a "draw text" command.
    // Each time a message/parameter/position command is received, its information
    // is stored in the cell for later use. When the draw command is received, the text
    // is drawn according to the currently stored info.
    // This means that message/parameter/position can be sent in any order, and
    // that you need to send them only if you want to change them
    Serial.println(F("Text 1 (red \"Hello\" on blue bg)..."));
    if(waitAnyKey()) { return; }
    commander.sendFill(yellow);  // just to show that this will be overwritten by the text's background!
    commander.sendText(txt.str);
    commander.sendTextPars(txt); // send everything *except* the message and the position
    commander.sendPosition(txt.topleft);
    commander.sendDrawText();    // actually draw the text, with the parameters given before
    commander.sendSwap();        // show the drawn text
    
    
    // Test 2
    // We change the parameters and draw the text again.
    txt.setText(F("1234"));
    txt.setColor(yellow);               // Color of the letters
    txt.setOverlay(true);               // Now the background is transparent
    txt.setBgColor(blue);               // The background color (will be ignored, since it's transparent!)
    txt.setMonospace(true);
    txt.setCharSpacing(0);              // being monospaced, there is already some spacing between not-too-wide characters
    
    Serial.println(F("Text 2 (yellow \"1234\" on violet bg, monospace)..."));
    if(waitAnyKey()) { return; }
    commander.sendFill(violet);  // prepare the background (overlay=true, so the text will be drawn over this background)
    commander.sendText(txt.str);
    commander.sendTextPars(txt); // send the new parameters
    //commander.sendPosition(txt.topleft); // no need to send this, we didn't chang it
    commander.sendDrawText();
    commander.sendSwap();
    
    
    // Test 3
    // We create a scrolling text
    txt.setText(F("FlashMat"));
    txt.setColor(green);
    txt.setOverlay(false);
    txt.setBgColor(black);
    txt.setMonospace(false);
    txt.setCharSpacing(1);
    txt.setLineSpacing(1);
    //txt.setTopleft(0,0);     // we'll set this dynamically
    
    // This function computes the pixels occupied by the text, depending
    // on the currently specified parameters (text, font, spacing, ...).
    Point size = txt.getStringPixelSize();
    
    Serial.println(F("Text 3 (scrolling green \"FlashMat\" on black bg)..."));
    if(waitAnyKey()) { return; }
    
    commander.sendText(txt.str);
    commander.sendTextPars(txt);
    
    for(coord_t x = MATRIX_COLS; x >= -size.x; --x) {
        // we loop along the x axis, to scroll the text from right to left
        commander.sendPosition(x, 0); // set the new position
        commander.sendDrawText();     // draw the text
        commander.sendSwap();         // show it
        delayMs(50);                  // wait some time to obtain the desired scrolling speed
    }
    
    
    // Test 4
    // A centered text that changes color
    txt.setText(F("Hello"));
    //txt.setColor(blue);      // we'll set this dynamically
    txt.setOverlay(false);
    txt.setBgColor(black);
    txt.setMonospace(false);
    txt.setCharSpacing(1);
    txt.setLineSpacing(1);
    //txt.setTopleft(0,0);     // we'll set this dynamically
    
    Serial.println(F("Text 4 (centered \"Hello\" blue->red->blue on black bg)..."));
    if(waitAnyKey()) { return; }
    
    // we prepare the common parameters
    commander.sendText(txt.str);
    commander.sendTextPars(txt);
    
    size = txt.getStringPixelSize();
    coord_t startX = (MATRIX_COLS - size.x) / 2; // center the text
    commander.sendPosition(startX, 0);
    
    // Loop: we gradually change color from blue to red
    for(int i = 0; i <= 255; i += 8) {
        uint8_t color[3] = MAKE_RGB(i, 0, 255-i);
        
        commander.sendTextPars(color);// All the parameters we don't send will be left unchanged; here we only need to change the color
        commander.sendDrawText();     // draw the text
        commander.sendSwap();         // show it
        delayMs(50);                  // wait some time to obtain the desired scrolling speed
    }
    
    // Now we go back to blue
    for(int i = 0; i <= 255; i += 8) {
        uint8_t color[3] = MAKE_RGB(255-i,0,i);
        
        commander.sendTextPars(color);
        commander.sendDrawText();
        commander.sendSwap();
        delayMs(50);
    }
    
    // Test 5
    // A text that slides in and out with an "animation"
    txt.setText(F(" Hello ")); // we use the leading and trailing spaces to "push" other letters away (see the loop below)
    txt.setColor(orange);
    txt.setOverlay(false);
    txt.setBgColor(black);
    txt.setMonospace(false);
    //txt.setCharSpacing(1);   // we'll set this dynamically
    txt.setLineSpacing(1);
    //txt.setTopleft(0,0);     // we'll set this dynamically
    
    Serial.println(F("Text 5 (slide-in orange \"Hello\" on black bg)..."));
    if(waitAnyKey()) { return; }
    
    // we prepare the common parameters
    commander.sendText(txt.str);
    commander.sendTextPars(txt);
    
    // we will make the text enter and exit by collapsing/expanding the space between letters.
    // To have the text outside the cell, we reach a quite big spacing
    const int maxSpace = MATRIX_COLS;
    
    // we compute how much space the "normal" text takes up
    size = txt.getStringPixelSize();
    
    // We center the text when it is "normal"
    startX = (MATRIX_COLS - size.x) / 2; // center the text
    coord_t stopX = startX + size.x - 1;         // the rightmost pixel (used when the text expands to the left)
    
    // Loop 1: we gradually collapse the text, as if it was entering from the right
    for(int s = maxSpace; s >= 1; --s) {
        // we change the text's spacing
        txt.setCharSpacing(s);
        
        commander.sendTextPars(txt);       // We only need to change the spacing, but the packet format forces us to resend most of the parameter
        commander.sendPosition(startX, 0); // The text always starts from here (it will converge here)
        commander.sendDrawText();
        commander.sendSwap();
        delayMs(50);
    }
    
    delayMs(250); // freeze the text for a little while
    
    // Loop 2: we gradually expand the text, as if it was leaving to the left
    for(int s = 1; s <= MATRIX_COLS; ++s) {
        // we change the text's spacing
        txt.setCharSpacing(s);
        
        Point size2 = txt.getStringPixelSize();
        
        commander.sendTextPars(txt);                // We only need to change the spacing, but the packet format forces us to resend most of the parameter
        commander.sendPosition(stopX - size2.x, 0); // We want the text to always end at stopX, so we must recalculate its starting position. (it will appear to diverge from stopX)
        commander.sendDrawText();
        commander.sendSwap();
        delayMs(50);
    }
}

/*
 * Sends the whole image as a bitmap.
 * Test for the packets:
 * + PKT_IMG_4bit_CHUNK
 * + PKT_SWAP
 */
void testBitmap() {
    printName("bitmap");
    
    /*
     * Since the memory on the cell is limited, we cannot send the whole bitmap
     * image at once (in this case, the specific reason is that the I2C buffer
     * that temporarily stores the image is not big enough).
     * So, we have to send it piece-by-piece (in "chunks"); the chunk is an 8x8-pixels
     * subimage.
     * Moreover, the cell has 4 bits of color depth (each pixel has 4 bits for
     * red, 4 for green and 4 for blue), so we speed up the communication by packing
     * two 4bit-colors in each byte.
     *
     * So, we can first create a full-sized, 8bit image and manipulate it; then, when
     * we must send it to the cell, we can compress and split it into suitable chunks.
     *
     * ==========================
     *
     * To be able to use the default functionalities, we must use this format for the array
     * containing the image: if the image has NR rows and NC cols, the image is a
     * uint8_t[3][NR][NC]. The red plane is img[R] (a uint8_t[NR][NC]), the green plane
     * is img[G] and the blue plane is img[B]. R,G,B are predefined macros that correspond
     * to the indices 0,1,2, for convenience.
     * Example: if we had a small 2x2 image:
     *   .========.========.
     *   | yellow | purple |
     *   |========|========|
     *   |  cyan  |  white |
     *   '========'========'
     * Its memory representation would be:
     *   img = [255, 255, 0, 255, 255, 0, 255, 255, 0, 255, 255, 255]  (as a linear array, with order RGB)
     * which is equivalent to:
     *   img[R] = [ 255, 255 ;
     *                0, 255 ]
     *   img[G] = [ 255,   0 ;
     *              255, 255 ]
     *   img[B] = [   0, 255 ;
     *              255, 255 ]
     *
     * If we compress it to 4bit-depth, we get (we ignore the least significant bits of each byte,
     * and group two bytes together):
     *   img[R] = [ 11111111, 11111111 ;   =>  1111 1111  = 255
     *              00000000, 11111111 ]   =>  0000 1111  =  15
     *   img[G] = [ 11111111, 00000000 ;   =>  1111 0000  = 240
     *              11111111, 11111111 ]   =>  1111 1111  = 255
     *   img[B] = [ 00000000, 11111111 ;   =>  0000 1111  =  15
     *              11111111, 11111111 ]   =>  1111 1111  = 255
     * which is equivalent to:
     *   img = [255, 15, 240, 255, 15, 255]
     *
     * ==========================
     *
     * When we have to split the image into 8x8 chunks, we must extract the proper bytes.
     * Example: NR = 8, NC = 16; the image is
     *    ================================ ================================
     *   | A0  A1  A2  A3  A4  A5  A6  A7 | A8  A9 A10 A11 A12 A13 A14 A15 |
     *   | B0  B1  B2  B3  B4  B5  B6  B7 | B8  B9 B10 B11 B12 B13 B14 B15 |
     *   | C0  C1  C2  C3  C4  C5  C6  C7 | C8  C9 C10 C11 C12 C13 C14 C15 |
     *   | D0  D1  D2  D3  D4  D5  D6  D7 | D8  D9 D10 D11 D12 D13 D14 D15 |
     *   | E0  E1  E2  E3  E4  E5  E6  E7 | E8  E9 E10 E11 E12 E13 E14 E15 |
     *   | F0  F1  F2  F3  F4  F5  F6  F7 | F8  F9 F10 F11 F12 F13 F14 F15 |
     *   | G0  G1  G2  G3  G4  G5  G6  G7 | G8  G9 G10 G11 G12 G13 G14 G15 |
     *    ================================ ================================
     *      ^ this is the chunk # (0,0)       ^ this is the chunk # (0,1)
     * (this shows one single color plane).
     * For each color plane, we identify an 8x8 chunk by a row number and a column
     * number. The topleft chunk is (r=0,c=0), the one to its right is (r=0,c=1),
     * the second "row" has chunks (r=1,c=0) and (r=1,c=1), and so on.
     * So, the chunk (A,B) spans the pixels (x=A*8,y=B*8)->(x=A*8+7,y=B*8+7).
     *
     * ==========================
     *
     * To send the bitmap to the cell, you must then:
     * + Convert the image to 4bit-depth
     * + Extract one chunk (A,B) at a time
     * + Send the chunk to the cell, specifying that it is the chunk number (A,B)
     *
     * Example: with NR = 8, NC = 32, you will send 4 chunks:
     * commander.sendImageChunk4bit(0,0,chunk1);  // note that the column number comes first!
     * commander.sendImageChunk4bit(1,0,chunk2);
     * commander.sendImageChunk4bit(2,0,chunk3);
     * commander.sendImageChunk4bit(3,0,chunk4);
     *
     * In this Arduino example code, the function Commander::sendImage4bitInChunks() takes care
     * of most of this work.
     */
    
    promptAnyKeyToStop();
    
    uint8_t img[3][MATRIX_ROWS][MATRIX_COLS]; // 8bit image (it's easier for us to draw it)
    uint8_t img2[FRAME_SIZE_FOR(4)];          // 4bit image (it's needed by the cell)
    
    commander.selectCell(address);
    
    int i = 0;
    for(;;) {
        if(Serial.available()) {
            return;
        }
        
        memset((uint8_t*)img, 0x00, FRAME_SIZE_FOR(8));      // black background
        drawLemming(img, i,true);                            // draw the i-th frame of the lemming animation
        if(i>=19) { drawLemming(img, i-19,true); }           // draw another lemming
        ImgDepthConverter::c8to4(img2, (const uint8_t*)img); // convert 8bit to 4bit
        commander.sendImage4bitInChunks(img2);               // split it in chunks and send every chunk
        commander.sendSwap();                                // show the whole image
        delay(80);
        
        i++;
        
        if(i < 0) { i = 0; } // for the unlikely case of overflow
    }
}

/*
 * Shows how to draw a consistent image over multiple cells, for which it is
 * known in advance where they are placed with respect to one another.
 */
void testMultiCell() {
    printName("multiCell");
    
    /*
     * To have many cells connected together showing a consistent image, we
     * must know where they will be placed and tell them where they are.
     * Here we assume that the cells are placed like this:
     *  +====+====+
     *  | 64 | 65 |
     *  +====+====+
     *  | 66 | 67 |
     *  +====+====+
     *  | 68 | 69 |
     *  +====+====+
     *  | 70 | 71 |
     *  +====+====+
     * (They will form a 32x64 pixels image)
     * To assign the proper addresses to each cell, see testCellAddress().
     */
    
    // For convenience, the array is BROADCAST-terminated (BROADCAST is not a valid address
    // for a cell)
    const uint8_t addresses[] = {64,65,66,67,68,69,70,71,BROADCAST};
    const uint8_t NC = 2;
    const uint8_t NR = 4;
    
    const uint8_t N = sizeof(addresses) / sizeof(*addresses) - 1;

    // We check that all the cells are there
    bool ok = true;
    for(const uint8_t* a = addresses; *a != BROADCAST; a++) {
        commander.selectCell(*a);
        if(commander.ping() != I2C_ERR_NONE) {
            Serial.print(F("Cell #"));
            Serial.print(int(*a));
            Serial.println(F(" not found"));
            ok = false;
        }
    }
    
    if(!ok) {
        Serial.println(F("Enter 'q' to quit, anything else to run the test anyway."));
        prompt2();
        if(waitAnyKey()) { return; }
    }
    
    /*
     * To tell each cell where it is with respect to the others, we define their
     * top-left pixel.
     */
    for(int i = 0; addresses[i] != BROADCAST; i++) {
        commander.selectCell(addresses[i]);
        // we assign
        // (x=0,y=0) for the first cell,
        // (x=C,y=0) for the second one,
        // (x=0,y=R) for the third one,
        // (x=C,y=R) for the fourth one, and so on
        commander.sendCellPosition(MATRIX_COLS * (i % NC), MATRIX_ROWS * (i / NC));
    }
        
    /*
     * We draw a gradient over the whole matrix: each cell will draw its part correctly.
     */
    // We send the same command to all the cells that we select via setBcastAddresses()
    // Note: you could also send the same command to all the single addresses in 'addresses';
    // you should not send them manually to the I2C address 0 (see testCellAddress(); this
    // address means "any connected cell"), because this was seen to cause problems of loss
    // of commands. Sending the command to all the cells singularly is safer.
    commander.setBcastAddresses(addresses);
    commander.selectCell(BROADCAST);
    commander.sendDrawGradient(violet, lime, GRAD_CIRCLE, Point(0,0), Point(NC * MATRIX_COLS - 1, NR * MATRIX_ROWS - 1));
    commander.sendSwap();
    
    delayMs(500);
    
    /*
     * Now we draw a gradient and a text on the first row, an horizontal rule on the second row
     * and a bigger text on the 3rd and 4th rows.
     * Since only the big text will be changing, we can refresh it on its own, while keeping
     * the other cells unchanged.
     */
    commander.selectCell(BROADCAST); // we will send commands to many cells: we'll change the addresses below when we need them
    
    if(N >= 2) {
        Text txt;
        
        // Draw the gradient and text on the first row
        txt.setText(F("FlashMat"));
        txt.setColor(blue);
        txt.setOverlay(true); // draw on the current background
        txt.setFont(Fonts::FM_8x8);
        txt.setMonospace(false);
        txt.setCharSpacing(1);
        Point size = txt.getStringPixelSize();
    
        const uint8_t addrs1[] = {addresses[0], addresses[1], BROADCAST};
            
        commander.setBcastAddresses(addrs1);
        commander.sendDrawGradient(yellow, red, GRAD_NORTH, Point(0,0), Point(MATRIX_COLS*NC - 1, MATRIX_ROWS - 1));
        commander.sendText(txt.str);
        commander.sendTextPars(txt);
        commander.sendPosition((MATRIX_COLS * NC - size.x) / 2, 0);
        commander.sendDrawText();
        commander.sendSwap(); // Note: it should be better to swap them together with all the other cells
                              // We do this here just for simplicity, since we don't need to swap them ever again
    }
    
    // Draw horizontal rule on the second row
    if(N >= 4) {
        const uint8_t addrs2[] = {addresses[2], addresses[3], BROADCAST};
        commander.setBcastAddresses(addrs2);
        commander.sendFill(black);
        commander.sendDrawRect(2, MATRIX_ROWS + 3, MATRIX_COLS * NC - 1 - 2, MATRIX_ROWS + 4, orange, true);
        commander.sendSwap(); // Note: as above
    }
    
    // Draw scrolling text on third and fourth row
    if(N >= 8) {
        Text txt;
        txt.setText(F("C&D Elettronica"));
        txt.setColor(white);
        txt.setOverlay(false);
        txt.setBgColor(black);
        txt.setFont(Fonts::ARIAL_16x16);
        txt.setMonospace(false);
        txt.setCharSpacing(2);
        
        const uint8_t addrs3[] = {addresses[4], addresses[5], addresses[6], addresses[7], BROADCAST};
        commander.setBcastAddresses(addrs3);
        
        commander.sendText(txt.str);
        commander.sendTextPars(txt);
        
        Point size = txt.getStringPixelSize();
        for(coord_t x = MATRIX_COLS*NC; x >= -size.x; x--) {
            if(Serial.available()) { return; }
            
            commander.sendPosition(x, MATRIX_ROWS * 2);
            commander.sendDrawText();
            commander.sendSwap();
            
            delay(50);
        }
    }
}

/*
 * Shows how to ensure proper drawing by waiting for a cell to complete its
 * job before sending another command.
 * This work is done automatically by standard Commander's functions, but
 * you can disable it and do it manually only when needed
 */
void testWait() {
    printName("wait");
    
    /*
     * When the commander sends a command, it can either be executed immediately (if it is fast)
     * or delayed. If a command is not delayed, the commander and the I2C bus will be
     * kept busy until the command has been completely processed.
     * Immediate commands are useful because when the next command arrives we are guaranteed
     * that the cell will be ready to process it.
     * Delayed commands are necessary to avoid keeping the commander busy (it might do something else
     * in the meantime, or send commands to different cells), but if a new command is sent to
     * the same cell while the old command is still being processed, the new command will be dropped
     * (probably causing an incorrect behavior).
     */
    
    Serial.println(F("You should see: solid violet; blue/yellow gradient; red/black gradient; white/black gradient"));
    
    commander.selectCell(address);
    
    commander.sendFill(violet, false); // no need to wait: fill is immediate
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC, false);
    // no waiting: since the swap is delayed, we should wait: we do it manually to show how
    commander.waitCell(address);
    
    delayMs(500);
    
    // We can also wait for multiple cells (see testMultiCell() for more info)
    uint8_t addrs[] = {address, BROADCAST}; // we can add as many addresses as we want, terminating them with BROADCAST
    commander.setBcastAddresses(addrs); // tell the commander which addresses we are interested in
    
    commander.sendDrawGradient(blue, yellow, GRAD_WEST,
                               Point(0,0), Point(MATRIX_COLS-1,MATRIX_ROWS-1),
                               false); // we should wait (gradient is delayed); we wait manually below
    commander.waitCells(); // we wait for all the addresses given to setBcastAddresses()
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC, true); // wait automatically
    
    delayMs(500);
    
    // Now we send a gradient and we *don't* wait: it will be drawn correctly,
    // but the next command will be dropped (the cell is still busy drawing the first one)
    // The swap comes too soon, too, but it is enqueued, and will be performed as soon as possible.
    // The swap command is the only one that can be enqueued; all other packets that arrive
    // when the cell is busy will be dropped.
    commander.sendDrawGradient(red, black, GRAD_WEST,
                                Point(0,0), Point(MATRIX_COLS-1,MATRIX_ROWS-1),
                                false); // we should wait (gradient is delayed); we wait manually below
    commander.sendDrawGradient(green, black, GRAD_WEST,
                                Point(0,0), Point(MATRIX_COLS-1,MATRIX_ROWS-1),
                                false); // we should wait (gradient is delayed); we wait manually below
    commander.sendSwap();
    
    // Note: these are the *delayed* commands, for which you must be careful to wait when needed:
    // PKT_SWAP
    // PKT_DRAW_TEXT
    // PKT_DRAW_LINE_H
    // PKT_DRAW_LINE_V
    // PKT_DRAW_RECT
    // PKT_DRAW_GRADIENT
    // PKT_DRAW_RAINBOW
    // PKT_STORE_ADDRESS   (for this, you should wait for the new address)
    
    /*
     * Now let's see how the commander knows when a cell is idle (ready for new commands)
     * or busy processing some command.
     * The commander starts a READ communication over the I2C (normally, it is a WRITE communication
     * to send data to the cell). The commander selects the cell address, and then it is up to the
     * cell to send back some data. The cell knows that when it is asked to, it must send
     * two bytes: its CellStatus and its PacketStatus.
     * The CellStatus can be either IDLE (the cell is idle) or BUSY (the cell is processing some command
     * and will drop any new command).
     * The PacketStatus can be ACCEPTED (the last packet was accepted), REJECTED (the last packet
     * was dropped because it was received while busy) or PENDING (the cell has just received the
     * packet and has not decided yet whether to keep it or drop it).
     *
     * This communication is done by Commander::reqStatus(). Based on the status you can decide
     * whatever you want: everything is ok, you have wait a bit more, you drop the operation and go on, ...
     *
     * Note: if a cell is not connected or has a serious problem, the returned values might be garbage:
     * they might not be one of the values IDLE/BUSY and ACCEPTED/REJECTED/PENDING.
     */
    
    // Send a command that will take some time to be processed
    commander.sendDrawGradient(white, black, GRAD_WEST,
                                Point(0,0), Point(MATRIX_COLS-1,MATRIX_ROWS-1),
                                false);
    
    delayMs(500);
    
    Serial.println();
    Serial.println(F("Any key to stop, in case of infinite loop (see the code for different wait options)"));
    
    // Manual wait
    for(;;) {
        std::pair<CellStatus, PacketStatus> status = commander.reqStatus();
        
        Serial.print(millis());
        Serial.print(F("ms : "));
        printStatus(status); // Print "CellStatus" and "PacketStatus" (print their code if they are non-standard, i.e. there was a communication error)
        Serial.println();
        
        // example 1: wait until the cell is idle again
        // This can hang forever if the cell is disconnected
        if(status.first == IDLE) { break; }
        
        // example 2: wait while the cell is busy (no wait if an error occurs)
        //if(status.first != BUSY) { break; }
        
        // example 3: retry on error
        if(status.second == REJECTED) {
            // ... wait for IDLE ...
            // ... resend the command (and maybe wait for it, or retry for a maximum number of times) ...
        }
        
        // Note: you should also stop the waiting if it goes on for too long, to
        // avoid infinite loops
        
        // Avoid endless loop just for this test
        if(Serial.available()) { break; }
    }
    
    // Now that we've waited, we can send the swap (maybe we should skip it if there was an error)
    commander.sendSwap();
}

/*
 * Shows how to check that all the desired cells are active and wait for them
 * to setup.
 */
void testStartup() {
    printName("startup");
    
    /*
     * At startup, the commander might be faster to power up and set up than the cells.
     * If the commander begins immediately to send commands (especially configuration commands
     * that will never be repeated later), the slower cells might miss them and then behave
     * incorrectly.
     *
     * In this test we just use one cell, with the usual address.
     *
     * During the other tests we didn't wait at startup, for convenience.
     * Since you must select a test via the serial, you can just wait some time yourself
     * before starting the first test after startup.
     */
    
    // We set the addresses we are interested in (see testMultiCell() for more info)
    uint8_t addrs[2] = {address, BROADCAST};
    commander.setBcastAddresses(addrs);
    
    const uint32_t msTimeout = 5000;
    
    Serial.print(F(
        "Now disconnect or reset the cell." "\n"
        "Then send any key over the serial and reconnect the cell." "\n"
        "From when you send the command, the cell must be powered up and be able to setup" "\n"
        "    (until it's ready to accept I2C commands) within "
    ));
    Serial.print(msTimeout);
    Serial.print(F(
        "ms" "\n"
        "Try this test twice: once reconnect the cell immediately;" "\n"
        "    the second time rerun this test and do not reconnect it: you'll get an error" "\n"
        "" "\n"
        "Enter a key when you're ready..."  "\n"
    ));
    prompt2();
    
    if(waitAnyKey()) { return; }
    i2c_err err = commander.waitForCellsToTurnOn(msTimeout); // This function simply pings the cells repeatedly until they're ready,
                                                             // or until the timeout is reached
    
    if(err == I2C_ERR_NONE) {
        Serial.println();
        Serial.println(F("... all cells were ok"));
    } else {
        Serial.println();
        Serial.println(F("... the cells were not ready in time"));
    }
}

/*
 * Shows the different types of swap.
 * Test for the packets:
 * + PKT_DRAW_RECT
 * + PKT_FILL
 * + PKT_SWAP
 */
void testSwapTypes() {
    printName("swapTypes");
    
    /*
     * To know the basics about swapping and double buffering, look at testFill().
     *
     * When we swap the front and back buffer, the new front buffer contains the image
     * to be shown, while the new back buffer is available to draw a new image.
     * If you redraw the complete image, what is left in the new back buffer
     * (which used to be the front buffer) is irrelevant. But if you only draw
     * some pixels of it (or lines, or text with transparent background, ...),
     * then the previous content of the back buffer will show through.
     * You can then control was is in the new back buffer right after the swap, before
     * you issue any drawing command.
     * If you provide SWAP_BLANK as a parameter to the swap command, the back buffer
     * will be blanked out (completely black); if you provide SWAP_NOBLANK, it will
     * contain the previously shown image (this is faster).
     *
     * Also, when we send a swap command, the cell will swap as soon as possible.
     * The cell also has a refresh cycle, during which it lights up one row of pixels
     * at a time. Since the commander cannot know which row is being lit up at one moment,
     * the swap can happen anytime during the refresh cycle; if the swaps happen at high rates,
     * this can cause annoying artifacts, like flickering.
     * To mitigate this you can provide SWAP_SYNC as a parameter to the swap command: the swap
     * will happen at the beginning of the next refresh cycle. The default parameter is SWAP_NOSYNC.
     * If you need high swap rates and see artifacts, try SWAP_SYNC, but also check that
     * it does not make things worse, since it can introduce other synchronization problems
     * (e.g. say you send a swap command at every refresh cycle, @100Hz; then each frame will
     * be shown for exactly one cycle; but if you send one command a bit late, then a new cycle
     * starts and the previous frame will be shown for 2 cycles, which is worse than having it
     * shown for 1-and-a-fraction cycle).
     */
    
    const uint16_t msDelay = 500;
    
    commander.selectCell(address);
    
    Serial.println(F("Any key to start each test case"));
    
    // Test 1
    Serial.println(F("No blank, no sync (red/green/red)..."));
    if(waitAnyKey()) { return; }
                                                       // front = ???,   back = ???
    commander.sendFill(red);                           // front = ???,   back = red
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = red,   back = ???
    delayMs(msDelay);
    commander.sendFill(green);                         // front = red,   back = green
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = green, back = red
    delayMs(msDelay);
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = red,   back = green
    delayMs(msDelay);
    
    // Test 2
    Serial.println(F("Blank, no sync (blue/yellow/black/yellow+rect)..."));
    if(waitAnyKey()) { return; }
                                                       // front = red,   back = green  (from previous test)
    commander.sendFill(blue);                          // front = red,   back = blue
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = blue,  back = red
    delayMs(msDelay);
    commander.sendFill(yellow);                        // front = blue,  back = yellow
    commander.sendSwap(SWAP_BLANK | SWAP_NOSYNC);      // front = yellow,  back = black
    delayMs(msDelay);
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = black,  back = yellow
    delayMs(msDelay);
    commander.sendDrawRect(2, 2, MATRIX_COLS - 3, MATRIX_ROWS - 3, violet, true);
                                                       // front = blue,  back = yellow+rect
    commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC);    // front = yellow+rect,  back = black
    delayMs(msDelay);
    
    // Test 3
    Serial.println(F("No blank, sync (no flickering)..."));
    if(waitAnyKey()) { return; }
    
    // We want a stroboscopic effect: we prepare a white image and a black one,
    // and then we swap them fast
    commander.sendFill(white);                      // white image
    commander.sendSwap(SWAP_BLANK | SWAP_NOSYNC);   // show the white image, and get a black one on the back buffer
    for(uint8_t i = 0; i < 101; i++) {              // swap the two images quite fast
        delayMs(30);
        commander.sendSwap(SWAP_NOBLANK | SWAP_SYNC);
    }
}

/*
 * Shows how to obtain an accurate frame rate.
 */
void testTiming() {
    printName("timing");
    commander.selectCell(address);
    
    /*
     * In all the other tests we used a delay of some milliseconds to
     * slow down the frames to a speed that is pleasing to the eye.
     * That is easy to do, but is far from perfect.
     * If you want to achieve an accurate frame rate, you cannot wait
     * for a fixed amount of time, because the total time of the frame is:
     *   total = delay
     *           + time to send the commands
     *           + time to process the commands(during which the commander may be waiting for the cell)
     *
     * These times are not constant, especially among commands of different complexity.
     * To wait for the proper amount of time, we must use a timer.
     */
    
    uint32_t T = 50; // ms
    
    uint32_t prevFrameMs = millis(); // we use this to remember the last time we send a swap command
                                     // (the instant at which the new frame is shown)
    
    for(int i = 0; i < 30; i++) {
        // some drawing command, just as example
        commander.sendFill((i%2) ? red : blue);
        
        bool safeTechnique = true;
        if(!safeTechnique) {
            delay(T);
        } else {
            while(millis() - prevFrameMs < T); // wait the exact time needed
            prevFrameMs = millis();
            // note: for extra precision you can use micros();
            // millis() can cause the frame rate to be off by +-1ms
        }
        
        // The wait must happen right before the swap: long computations
        // between the wait and the swap would cause timing errors
        commander.sendSwap();
    }
    
    /*
     * Note that if the drawing commands are slow, you might get a slower frame rate than desired.
     * Anyway, the problem is more evident when using delays than when waiting for the timer
     * to hit a desired deadline.
     */
    
    for(int j = 0; j < 2; j++) {
        // We experiment with two frame rates: one that is ok and one that
        // is too fast. The delay technique is worse in both cases.
        T = (j == 0) ? 100 : 10;
        Serial.println();
        Serial.print(F("Desired frame time = "));
        Serial.print(T);
        Serial.println(F("ms"));
    
        Serial.print(F("Actual frame time with delay ="));
        for(int i = 0; i < 10; i++) {
            const uint32_t t0 = millis();
            // some drawing command, just as example
            commander.sendDrawGradient((i%2) ? red : blue, yellow, GRAD_NORTH_WEST,
                                   Point(0,0), Point(MATRIX_COLS-1, MATRIX_COLS-1));
        
            delay(T);
        
            Serial.print(' ');
            Serial.print(millis() - t0);
        
            commander.sendSwap();
        }
        Serial.println();
    
        Serial.print(F("Actual frame time with timer ="));
        for(int i = 0; i < 10; i++) {
            const uint32_t t0 = millis();
            // some drawing command, just as example
            commander.sendDrawGradient((i%2) ? red : blue, yellow, GRAD_NORTH_WEST,
                                   Point(0,0), Point(MATRIX_COLS-1, MATRIX_COLS-1));
        
            while(millis() - t0 < T);
        
            Serial.print(' ');
            Serial.print(millis() - t0);
        
            commander.sendSwap();
        }
        Serial.println();
    }
}

void drawLemming(uint8_t img[3][MATRIX_ROWS][MATRIX_COLS], int frame, bool overlay) {
    static const uint8_t N_FRAMES = 8;
    static const uint8_t H = 8; // in pixels
    static const uint8_t W = 6; // actual non-empty width in pixels
    
    // H is the height (in pixels and rows)
    // W is the width (in pixels)
    // bytesW is the displacement between consecutive rows in the frame mask
    const uint8_t bytesW = ((W + 7) / 8) * 2; // width in bytes (rounded to ceiling) then *2 since each pixel takes 2bits (to index 4 colors)
    
    // Each frame is encoded using indices 0~3 for 4 colors:
    //   0 = background
    //   1 = clothes
    //   2 = hair
    //   3 = skin
    // The bits are packed, so that 1 byte contains 4 2-bit indices;
    //  the MSBs are for the leftmost pixel: the image can be "read" left-to-right.
    // The rightmost pixels might be all background (they may be needed in order
    //  to have an integer number of bytes per row)
    static const uint8_t walker_frames[8][16] PROGMEM = {
		{ // Frame 0
			0x2A,0x80, // . & & & & .
			0x2B,0x00, // . & & = . .
			0x0F,0xC0, // . . = = = .
			0x0D,0x00, // . . = # . .
			0x0D,0x00, // . . = # . .
			0x05,0x00, // . . # # . .
			0x35,0x00, // . = # # . .
			0x0F,0x00  // . . = = . .
		},
		{ // Frame 1
			0x2A,0x00, // . & & & . .
			0x2B,0x00, // . & & = . .
			0x0F,0xC0, // . . = = = .
			0x0D,0x00, // . . = # . .
			0x35,0x00, // . = # # . .
			0x35,0x30, // . = # # . =
			0x14,0xC0, // . # # . = .
			0x3C,0x00  // . = = . . .
		},
		{ // Frame 2
			0x22,0x00, // . & . & . .
			0x2A,0x00, // . & & & . .
			0x0B,0x00, // . . & = . .
			0x0F,0xC0, // . . = = = .
			0x3D,0x00, // . = = # . .
			0xF5,0x00, // = = # # . .
			0x15,0x40, // . # # # # .
			0xF0,0xF0  // = = . . = =
		},
		{ // Frame 3
			0x0A,0x00, // . . & & . .
			0x2B,0x80, // . & & = & .
			0x2F,0xC0, // . & = = = .
			0x0D,0x00, // . . = # . .
			0x0D,0x00, // . . = # . .
			0x35,0x00, // . = # # . .
			0xD5,0x40, // = # # # # .
			0xC3,0xC0  // = . . = = .
		},
		{ // Frame 4
			0x2A,0x80, // . & & & & .
			0x2B,0x00, // . & & = . .
			0x2F,0xC0, // . & = = = .
			0x0D,0x00, // . . = # . .
			0x07,0x00, // . . # = . .
			0x0D,0x00, // . . = # . .
			0x35,0x00, // . = # # . .
			0x0F,0x00  // . . = = . .
		},
		{ // Frame 5
			0x2A,0x00, // . & & & . .
			0x2B,0x00, // . & & = . .
			0x0F,0xC0, // . . = = = .
			0x0D,0x00, // . . = # . .
			0x07,0x00, // . . # = . .
			0x05,0x30, // . . # # . =
			0x14,0xC0, // . # # . = .
			0x3C,0x00  // . = = . . .
		},
		{ // Frame 6
			0x22,0x00, // . & . & . .
			0x2A,0x00, // . & & & . .
			0x0B,0x00, // . . & = . .
			0x0F,0xC0, // . . = = = .
			0x07,0x00, // . . # = . .
			0x05,0xC0, // . . # # = .
			0x15,0x40, // . # # # # .
			0xF0,0xF0  // = = . . = =
		},
		{ // Frame 7
			0x0A,0x00, // . . & & . .
			0x2B,0x80, // . & & = & .
			0x2F,0xC0, // . & = = = .
			0x0D,0x00, // . . = # . .
			0x0D,0x00, // . . = # . .
			0x07,0x00, // . . # = . .
			0xD5,0x40, // = # # # # .
			0xC3,0xC0  // = . . = = .
		}
	};

    static const uint8_t colors[4][3] = { MAKE_RGB(0,0,0), // black bg
                                          MAKE_RGB(0,0,255), // blue clothes
                                          MAKE_RGB(0,255,0), // green hair
                                          MAKE_RGB(255,235,215), // skin
                                        };
    
    // calculate frame number and position
    coord_t x = (frame % (MATRIX_COLS + W)) - W;
    frame = frame % N_FRAMES;
    
    // fill with background color
    if(!overlay) {
        memset(&img[R][0][0], colors[0][R], FRAME_SIZE_1_COLOR_FOR(8));
        memset(&img[G][0][0], colors[0][G], FRAME_SIZE_1_COLOR_FOR(8));
        memset(&img[B][0][0], colors[0][B], FRAME_SIZE_1_COLOR_FOR(8));
    }
    
    if(x <= -W || x >= MATRIX_COLS) {
        // nothing to draw:
        return;
    }
    
    const uint8_t* wlkFrame = walker_frames[frame];
    // draw the image in the right place
    // We only iterate over the pixels that might be colored; the other ones
    // have already been filled as background
    for(uint8_t yy = 0; yy < H; yy++) { // iterate over rows
        coord_t xp = x;
        for(uint8_t xx = 0; xx < W;) { // iterate over columns
            // optimization: to execute only the minimum number of pgm_read_byte
            // (which are slow), we read one byte and iterate over its 4 groups of 2 bits
            // (each group identifies one of 4 colors)
            uint8_t val = pgm_read_byte(&wlkFrame[xx / 4 + yy * bytesW]);
            
            for(uint8_t i = 0; i < 4 && xx < W; i++, xx++, xp++) {
                if(xp < 0 || xp >= MATRIX_COLS) continue; // too far to the left/right: draw nothing
                
                uint8_t index = (val >> (6 - (2*i))) & 0x03;
                
                if(index != 0 || true) { //skip background color
                    // drawPixel
                    img[R][yy][xp] = colors[index][R];
                    img[G][yy][xp] = colors[index][G];
                    img[B][yy][xp] = colors[index][B];
                }
            }
        }
    }
}

/*
 * A nice utility to know which cell has which address.
 * Checks which address are responsive, and then draws a text on the cell,
 * showing the address.
 */
void testIdentify() {
    printName("identify");
    
    static bool odd = true;
    odd = !odd;

    Text txt;
    txt.setColor(odd ? blue : red);
    txt.setOverlay(false);
    txt.setBgColor(black);
    txt.setFont(Fonts::FM_8x8);
    txt.setMonospace(false);
    txt.setCharSpacing(1);
    
    Serial.println(F("Found:"));
    
    for(uint8_t a = 8; a <= 119; a++) {
        commander.selectCell(a);
        if(commander.ping() == I2C_ERR_NONE) {
            Serial.print('\t');
            Serial.println(int(a));
            
            // The cell exists: draw its address on it
            itoa(a, txt.str, 10);            // convert number to string
            Point size = txt.getStringPixelSize();
            coord_t x0 = (MATRIX_COLS - size.x) / 2;
            
            commander.sendCellPosition(0,0); // to override settings from previous tests
            commander.sendText(txt.str);
            commander.sendTextPars(txt);
            commander.sendPosition(x0,0);    // center text
            commander.sendDrawText();
            commander.sendSwap();
        }
    }
}

/*
 * Just a simple way to turn all the cells off if they are too bright and
 * annoying.
 */
void blankAll() {
    printName("blankAll");
    
    for(uint8_t addr = 8; addr <= 119; addr++) {
        commander.selectCell(addr);
        if(commander.ping() == I2C_ERR_NONE) {
            commander.sendFill(black, false);
            commander.sendSwap(SWAP_NOBLANK | SWAP_NOSYNC, false);
        }
    }
}
