/*
 * Example to use a FlashMat master to create a wall clock using an RTC module.
 * 4 FlashMat cells are composed in a 2x2 grid, to create a 16x64-pixel display.
 *
 * For Arduino Nano
 *   RTC (DS1302)
 *       clock -> pin 2
 *       data  -> pin 3
 *       reset -> pin 4
 *       VCC   -> pin 5
 * BUTTON_UP   -> pin 7
 * BUTTON_DOWN -> pin 8
 *       SCL   -> pin A5
 *       SDA   -> pin A4
 *
 * The buttons are normally-open and pulled up to Vcc. The non-pulled-up terminal is connected
 * to the Arduino.
 *
 * Hold both buttons to enter/exit the set mode. Press a single button to set the time or
 * to change the display color.
 *
 * Note: you need the DS1302 library from http://www.henningkarlsen.com.
 * Note: if you experience strange behavior from the buttons, add a small capacitor between
 * ground and the switch's terminal that's connected to Arduino, to get a "debounce" effect.
 */

/*
 Copyright (C) 2013 C&D Elettronica.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
   
#include "FlashMat_master.h"

#include "DS1302.h"
#include <avr/eeprom.h>

using namespace FlashMat;

/******** CONSTANTS DECLARATIONS *********/
const uint8_t CELL_ADDRESS[2][2] = {{64, 65},{66, 67}};     // Address of the FlashMat cells
const int RTC_VCC     = 5;
const int RTC_DATA    = 3;
const int RTC_RESET   = 4;
const int RTC_CLOCK   = 2;
const int BUTTON_UP   = 7;
const int BUTTON_DOWN = 8;
const int PRESSED   = 0; // Buttons are pressed when digitalRead returns low (they are pull-up, normally open switches)
const int RELEASED  = 1;
const uint16_t CFG_PRESSED_TIME     = 1000; // Time [ms] to enter config mode
const uint16_t BLINK_TIME           =  300; // Delay between blinks (in config mode)
const uint16_t LOOP_PERIOD          =   50; // Time between successive readings of the sensors and buttons [ms]
const uint16_t T_PERIOD             =  500; // Time between successive readings of the time [ms]

const bool IGNORE_BAD_READINGS = false; // Ignore bad readings from RTC and temp sensor (disconnected, broken, ...) and use the last valid value instead
const bool DRAW_LEADING_ZERO = true; // Draw also the first '0' for hours < 10

const uint8_t NUM_COLORS = 10;
uint8_t fg_color = 0;

uint8_t COLOR_FG[NUM_COLORS][3] = { MAKE_RGB(255,   0,   0),
                                    MAKE_RGB(128,   0,   0),
                                    MAKE_RGB(  0, 255,   0),
                                    MAKE_RGB(  0, 128,   0),
                                    MAKE_RGB(  0,   0, 255),
                                    MAKE_RGB(  0,   0, 128),
                                    MAKE_RGB(255, 223,   0),
                                    MAKE_RGB(128, 112,   0),
                                    MAKE_RGB(255, 255, 255),
                                    MAKE_RGB(128, 128, 128)
                                  };
uint8_t COLOR_BG[3] = MAKE_RGB(0, 0, 0);

// Standard width and spacing of digits (to align them regardless of their actual size)
const int8_t DIGIT_WIDTH = 10;
const int8_t DIGIT_SPACING = 2;
const int8_t COLON_SPACING = 3;

/********* GLOBAL VARIABLES *********/

DS1302 rtc(RTC_RESET, RTC_DATA, RTC_CLOCK); // Instantiate a new object DS1302
Time time; // Allocate a support struct to save the current time
bool btnUp = RELEASED; // To save the current state of the button
bool btnDown = RELEASED;
bool prevBtnUp = RELEASED;
bool prevBtnDown = RELEASED;
bool waitBtnRelease = true; // buttons just pressed together, must wait for them to be released before detecting further presses
uint32_t lastUpPress = 0; // to know how long we've been pressing each button
uint32_t lastDownPress = 0;
Text txt;

uint8_t digitWidth[12]; // pixels of width of each digit
                        // [10] is the dot
                        // [11] is the colon

// To know what we must show on the matrix at each moment
enum State {ST_CLOCK, ST_CFG_HH, ST_CFG_MM};
State state = ST_CLOCK;
State prevState = ST_CFG_HH; // To know when we have just changed state

/********* FUNCTION PROTOTYPES *********/
void initText();
void readTime();
void readButtons();
bool checkConfigMode(uint16_t pressTime);
int  checkUpDown();
void setTime(const Time& time);
void drawClock(bool showHour = true, bool showMin = true, bool blinkColon = true);
void state_clock();
void state_cfg_hh();
void state_cfg_mm();
void state_cfg_helper(bool cfgHour);


/********* FUNCTION DEFINITIONS *********/

void setup() {
    pinMode(BUTTON_UP, INPUT_PULLUP);   // Set software-pullup resistor for two buttons
    pinMode(BUTTON_DOWN, INPUT_PULLUP);
    pinMode(RTC_VCC, OUTPUT);
    digitalWrite(RTC_VCC, HIGH);
    
    fg_color = eeprom_read_byte((const uint8_t*)0);
    if(fg_color >= NUM_COLORS) {
        fg_color = 0;
    }
    
    initText();
    
    rtc.writeProtect(false);  // Enable to change the RTC internal time
    rtc.halt(false);          // Start the RTC's internal counting (especially for the first time we use it)
    
    // Initial readings
    readTime();
    readButtons();
    
    /* Flashmat and I2C setup (for explanation, see the "Flashmat_master_example") */
    commander.setup();
    commander.setTimeoutUs(5000);
    commander.setPacketDelayUs(0);
    commander.setNumRetry(1);
    commander.setWaitTimeoutMs(200);
    delay(1000); //Give some time to the cells to power up and set up
    I2C.setFreq(400000L);
    
    // Wait for the cell to respond
    for(uint8_t r = 0; r < 2; r++) {
        for(uint8_t c = 0; c < 2; c++) {
            commander.selectCell(CELL_ADDRESS[r][c]);
            while(commander.ping() != I2C_ERR_NONE) {
                delay(250);
            }
            
            commander.sendCellPosition(32 * c, 8 * r);
        }
    }
    
    // one-time cell configurations
    uint8_t addrs[5] = { CELL_ADDRESS[0][0],
                         CELL_ADDRESS[0][1],
                         CELL_ADDRESS[1][0],
                         CELL_ADDRESS[1][1],
                         BROADCAST};
    commander.setBcastAddresses(addrs);
    commander.selectCell(BROADCAST);
    commander.sendTextPars(txt);
}

void initText() {
    txt.setFont(Fonts::ARIAL_16x16);
    txt.setColor(COLOR_FG[fg_color]);
    txt.setOverlay(true);
    txt.setBgColor(COLOR_BG);
    txt.setMonospace(false);
    txt.setCharSpacing(1);
    txt.setLineSpacing(1);
    txt.setTopleft(0, 0);
    
    char digits[12] = {'0','1','2','3','4','5','6','7','8','9','.',':'};
    for(int i = 0; i < 12; i++) {
        const char str[2] = {digits[i], '\0'};
        txt.setText(str);
        digitWidth[i] = txt.getStringPixelSize().x;
    }
    txt.setText("");
}

void loop() {
    static uint32_t t = millis(); // for the button reading and refresh period
    static uint32_t tt = t; // for the rtc and temp reading
    
    const uint32_t now = millis();
    
    if(now - tt >= T_PERIOD) {
        if(state != ST_CFG_HH && state != ST_CFG_MM) {
            // read only when not configuring, or the configuration will be overwritten
            readTime();
        }
        tt = now;
    }
    
    // execute the loop only once per period
    if(now - t < LOOP_PERIOD) {
        return;
    }
    t = now;
        
    // Read inputs
    readButtons();
    
    // State machine to take appropriate action for each state
    switch(state) {
        case ST_CLOCK:
            state_clock();
            prevState = ST_CLOCK;
            break;
        case ST_CFG_MM:
            state_cfg_mm();
            prevState = ST_CFG_MM;
            break;
        case ST_CFG_HH:
            state_cfg_hh();
            prevState = ST_CFG_HH;
            break;
        default: // should never happen!
            prevState = state;
            state = ST_CLOCK;
            break;
    }
}

void readButtons() {
    uint32_t now = millis();
    bool newBtnUp = digitalRead(BUTTON_UP);
    bool newBtnDown = digitalRead(BUTTON_DOWN);
    
    if(newBtnUp == PRESSED && btnUp != PRESSED) {
        lastUpPress = now;
    }
    
    if(newBtnDown == PRESSED && btnDown != PRESSED) {
        lastDownPress = now;
    }
    
    prevBtnUp = btnUp;
    prevBtnDown = btnDown;
    
    btnUp = newBtnUp;
    btnDown = newBtnDown;
}

void readTime() {
    const int numRetry = 3;
    for(int i = 0; i < numRetry; i++) {
        digitalWrite(RTC_VCC, LOW); // to avoid bad readings
        Time newT = rtc.getTime();
        digitalWrite(RTC_VCC, HIGH);
    
        if(IGNORE_BAD_READINGS ||
                    (newT.sec != 0 || newT.min != 0 || newT.hour != 0
                    || newT.date != 0 || newT.mon != 0 || newT.dow != 0 || newT.year != 2000)) {
            // This is a very simplified way to test if we read the RTC correctly.
            // If not, we just do not update the time.
            time = newT;
            return;
        }
    }
}

void state_clock() {
    // Each time we enter this state from another one, reset the counter
    // we use to skip to the next state
    const uint32_t now = millis();
    static uint32_t t = now;
    
    if(prevState != ST_CLOCK) {
        t = now;
    }
    
    drawClock();
    
    // Change display
    if(checkConfigMode(CFG_PRESSED_TIME)) {
        state = ST_CFG_HH;
    } else {
        int var = checkUpDown();
        
        if(var != 0) {
            fg_color = (fg_color + NUM_COLORS + var) % NUM_COLORS;
            eeprom_write_byte((uint8_t*) 0, fg_color);
            txt.setColor(COLOR_FG[fg_color]);
            commander.sendTextPars(txt);
        }
    }
}

void state_cfg_helper(bool cfgHour) {
    // Each time we enter this state from another one, reset the counter
    // we use to skip to the next state
    const uint32_t now = millis();
    static uint32_t t = now;
    static bool blinkOn = false;

    if(prevState != (cfgHour ? ST_CFG_HH : ST_CFG_MM)) {
        t = now;
        blinkOn = false;
    }
    
    int step = checkUpDown();
    // increase or decrease hour
    if(cfgHour) {
        if(step > 0) {
            time.hour = (time.hour == 23) ? 0 : time.hour+1;
        } else if (step < 0) {
            time.hour = (time.hour == 0) ? 23 : time.hour-1;
        }
    } else {
        if(step > 0) {
            time.min = (time.min == 59) ? 0 : time.min+1;
        } else if (step < 0) {
            time.min = (time.min == 0) ? 59 : time.min-1;
        }
    }
    
    if(step != 0) {
        setTime(time);
    }

    if(now - t >= BLINK_TIME) {
        t = now;
        blinkOn = !blinkOn;
    }
    
    if(step != 0) {
        // always show the time after a change (reset the blink)
        t = now;
        blinkOn = true;
    }

    if(cfgHour) {
        drawClock(blinkOn, true, false);
    } else {
        drawClock(true, blinkOn, false);
    }
    
    // Change display
    if(checkConfigMode(0)) {
        state = (cfgHour ? ST_CFG_MM : ST_CLOCK);
    }
    
}

void state_cfg_hh() {
    state_cfg_helper(true);
}

void state_cfg_mm() {
    state_cfg_helper(false);
}

void drawClock(bool showHour, bool showMin, bool blinkColon) {
    uint8_t h1 = uint8_t((uint8_t)time.hour / 10);    // extract 10's digit
    uint8_t h2 = uint8_t((uint8_t)time.hour - h1*10); // extract 1's digit
    uint8_t m1 = uint8_t((uint8_t)time.min / 10);     // extract 10's digit
    uint8_t m2 = uint8_t((uint8_t)time.min - m1*10);  // extract 1's digit
    
    const int8_t colonPos = MATRIX_COLS - 1;
    
    // draw background
    commander.sendFill(COLOR_BG);
    
    // draw ':'
    if(!blinkColon || millis() % 2000 < 1000) {
        commander.sendText(":");
        commander.sendPosition(colonPos, 0);
        commander.sendDrawText();
    }
    
    // draw hour
    if(showHour) {
        coord_t x1 = colonPos - 1 - COLON_SPACING - DIGIT_SPACING - DIGIT_WIDTH - (digitWidth[h1] + DIGIT_WIDTH) / 2;
        coord_t x2 = colonPos - 1 - COLON_SPACING - (digitWidth[h2] + DIGIT_WIDTH) / 2;
        
        if(DRAW_LEADING_ZERO || h1 != 0) {
            const char str1[2] = {char(h1 + '0'), '\0'};
            commander.sendText(str1);
            commander.sendPosition(x1, 0);
            commander.sendDrawText();
        }
        
        const char str2[2] = {char(h2 + '0'), '\0'};
        commander.sendText(str2);
        commander.sendPosition(x2, 0);
        commander.sendDrawText();
    }
    
    // draw min
    if(showMin) {
        coord_t x1 = colonPos + digitWidth[11] + COLON_SPACING + DIGIT_WIDTH - (digitWidth[m1] + DIGIT_WIDTH) / 2;
        coord_t x2 = colonPos + digitWidth[11] + COLON_SPACING + DIGIT_SPACING + 2*DIGIT_WIDTH - (digitWidth[m2] + DIGIT_WIDTH) / 2;
        
        const char str1[2] = {char(m1 + '0'), '\0'};
        commander.sendText(str1);
        commander.sendPosition(x1, 0);
        commander.sendDrawText();
        
        const char str2[2] = {char(m2 + '0'), '\0'};
        commander.sendText(str2);
        commander.sendPosition(x2, 0);
        commander.sendDrawText();
    }
    
    // Show everything
    commander.sendSwap();
}

/* Check if the two buttons have been pressed simultaneously for long enough to
 * enter/exit/skip config mode
 */
bool checkConfigMode(uint16_t minTime) {
    if(waitBtnRelease) {
        waitBtnRelease = (btnUp == PRESSED || btnDown == PRESSED);
    } else {
        uint32_t now = millis();
        if(btnUp == PRESSED && btnDown == PRESSED
                && now - lastUpPress >= minTime && now - lastDownPress >= minTime) {
            waitBtnRelease = true;
            return true;
        }
    }
    return false;
}

/* Check if a button up/down has been pressed and released (we check it has been released
 * to avoid conflict with the non-perfectly-simultaneous press of both buttons when
 * entering/exiting config mode)
 * Return -1, 0 or 1
 */
int checkUpDown() {
    if(waitBtnRelease) {
        waitBtnRelease = (btnUp == PRESSED || btnDown == PRESSED);
        return 0;
    }
    
    int step = 0;
    
    if(btnUp == RELEASED && prevBtnUp == PRESSED) {
        step++;
    }
    
    if(btnDown == RELEASED && prevBtnDown == PRESSED) {
        step--;
    }
    
    return step;
}

/* Write the currently shown time to the RTC */
void setTime(const Time& time) {
    rtc.setTime(time.hour, time.min, time.sec);
}