/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SERIAL_PARSE_H_
#define SERIAL_PARSE_H_

/*
 * Utilities to read a number from the serial, much faster than the Arduino's default one.
 * Note this, however:
 * + before calling this, make sure that all the characters have arrived (e.g. wait
 *   a reasonable amount of time, like some tens of ms), because the parsing stops
 *   as soon as the serial buffer is empty
 * + the parsing stops at the first wrong character (non-digit, or + - . not in the right place,...)
 *   returning the number read so far
 * + if no number is parsed at all, 0 is returned. You can distinguish this from a proper 0 by
 *   providing the 'parsed' argument, which is set to true in the case of a proper 0.
 * + you don't need to precede this call with a call to Serial.available(), but if you do that you
 *   then don't need to check for 'parsed'
 * + no overflow check is performed on the parsed number!
 */

#include <Arduino.h>
#include <inttypes.h>

inline double parseDouble(Stream& stream, bool* parsed = 0 /*null*/) {
    bool isNegative = false;
    bool isFraction = false;
    long value = 0;
    int16_t c;
    double fraction = 1.0;
    
    
    c = stream.peek();
    // ignore non numeric leading characters
    if((c < '0' && c > '9') && !(c== '-' || c== '+' || c=='.')){
        if(parsed)
            *parsed = false;
        return 0;
    }

    if(parsed)
        *parsed = true;

    if(c == '+'){
        isNegative = false;
        stream.read();
    } else if(c == '-') {
        isNegative = true;
        stream.read();
    }

    for(;;) {
        c = stream.peek();
        if (c == '.'){
            if(isFraction)
                break; // cannot have multiple dots
            isFraction = true;
        } else if(c >= '0' && c <= '9') {      // is c a digit?
            value = value * 10 + (c - '0');
            if(isFraction) {
                fraction *= 0.1;
            }
        } else {
            break;
        }
        stream.read();  // consume the character we got with peek
    }

    if(isNegative)
        value = -value;
    
    if(isFraction){
        return value * fraction;
    } else {
        return value;
    }
}

inline long parseLong(Stream& stream, bool* parsed = 0) {
    boolean isNegative = false;
    long value = 0;
    int16_t c;

    c = stream.peek();
    // ignore non numeric leading characters
    if((c < '0' && c > '9') && !(c== '-' || c== '+')){
        if(parsed)
            *parsed = false;
        return 0;
    }
    if(parsed)
        *parsed = true;

    if(c == '+') {
        isNegative = false;
        stream.read();
    } else if(c == '-') {
        isNegative = true;
        stream.read();
    }

    for(;;) {
        c = stream.peek();
        if(c >= '0' && c <= '9')  {      // is c a digit?
            long nValue = value * 10 + (c - '0');
            if(nValue < value) { // overflow
                if(parsed)
                    *parsed = true;
                return 0;
            }
            value = nValue;
        } else {
            break;
        }
        stream.read();  // consume the character we got with peek
    }

    if(isNegative)
        value = -value;
        
    return value;
}

#endif /* SERIAL_PARSE_H_ */
