/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SERIALIZER_H_
#define SERIALIZER_H_

#include <inttypes.h>

/*
 * Simple functions to split various data types into bytes and to
 * rejoin them.
 * Can slice any type of data/struct/class/enum/union.
 * (for implementation choices, the data structure can be no larger than
 *  2^32-1 bytes)
 * NOTE: the (de)serialization is made byte by byte, so the pointer variables
 * inside the objects you provide are not dereferenced (only the stack-allocated
 * bytes are copied).
 * Complex data types are intantiated using the default constructor, then
 * reconstructed byte-by-byte and finally returned by copy.
 *
 * You can use deserializeR to avoid this: instead of returning the rebuilt object,
 * the function takes as argument an object you already allocated, and modifies it.
 *
 * Serializer functions take an array of bytes where they put the sliced result;
 * you must guarantee there's enough room.
 * Deserializer functions take an array where they expect to find the sliced data
 * and return the joined data.
 *
 * A data D=[BN ... B1 B0] will be sliced into:
 * [BN ... B1 B0]  if 'msbFirst' == true
 * [B0 B1 ... BN]  if 'msbFirst' == false
 *
 * Usage examples:
 *    uint8_t buff[10];
 *    serialize<int>(73, buff);
 *    serialize<float>(-2.4f, buff + sizeof(int), false);
 *    int i = deserialize<int>(buff);
 *    float f = deserialize<float>(buff + sizeof(int), false);
 *  ==================
 *    uint8_t buff[10];
 *    struct MyStruct{bool b; char c};
 *    MyStruct ms = {true, 'a'};
 *    serialize<MyStruct>(ms, buff);
 *    MyStruct ms2 = deserialize<MyStruct>(buff);
 *  ==================
 *    uint8_t buff[100];
 *    class MyClass{...};
 *    MyClass mc(...);
 *    serialize<MyClass>(mc);
 *    MyClass mc2(...); // default ctor, unless it's not accessible: anyway the object
 *                      // will be completely overwritten.
 *                      // for the heap, you could also malloc(sizeof(MyClass)), to skip
 *                      // the ctor execution
 *    deserializeR(mc2, buff);
 *   ==================
 *
 * Note that for each different type used in calls to serialize() etc. a new function
 * is allocated, as it always is with templates.
 * (this ha nothing to do with the functions being declared inline)
 *
 * You might also redefine these functions for some of your datatypes, to make them
 * work with pointers, non-persistent data, ...; the caller wouldn't know that, and
 * could just call serialize<MyType> as before.
 *
 * If you want to avoid the code duplication due to templates (or you are restricted
 * to plain C), you can use the serializeV() and deserializeRV() functions.
 * They all take void*, and you must specify the size of your object.
 * deserializeV() doesn't exist, as it would need a copy operator (copy-ctor, assignment)
 * to return the built object, but the object type is unknown.
 */
template<typename T>
inline void serialize(const T& data, uint8_t* buffer, bool msbFirst = true) {
    const int16_t N = sizeof(T);
    const uint8_t* d = reinterpret_cast<const uint8_t*>(&data);

    if (msbFirst) {
        for (int16_t i = N; --i >= 0;) {
            *buffer++ = *d++;
        }
    } else {
        buffer += N - 1;
        for (int16_t i = N; --i >= 0;) {
            *buffer-- = *d++;
        }
    }
}

template<typename T>
inline void deserializeR(T& data, const uint8_t* buffer, bool msbFirst = true) {
    const int16_t N = sizeof(T);
    uint8_t* d = reinterpret_cast<uint8_t*>(&data);

    if (msbFirst) {
        for (int16_t i = N; --i >= 0;) {
            *d++ = *buffer++;
        }
    } else {
        buffer += N - 1;
        for (int16_t i = N; --i >= 0;) {
            *d++ = *buffer--;
        }
    }
}

template<typename T>
inline T deserialize(const uint8_t* buffer, bool msbFirst = true) {
    T data;
    deserializeR<T>(data, buffer, msbFirst);
    return data;
}

inline void serializeV(const void* data, uint16_t size, uint8_t* buffer,
        bool msbFirst = true) {
    const int16_t N = size;
    const uint8_t* d = reinterpret_cast<const uint8_t*>(data);

    if (msbFirst) {
        for (int16_t i = N; --i >= 0;) {
            *buffer++ = *d++;
        }
    } else {
        buffer += N - 1;
        for (int16_t i = N; --i >= 0;) {
            *buffer-- = *d++;
        }
    }
}

inline void deserializeRV(void* data, uint16_t size, const uint8_t* buffer,
        bool msbFirst = true) {
    const int16_t N = size;
    uint8_t* d = reinterpret_cast<uint8_t*>(data);

    if (msbFirst) {
        for (int16_t i = N; --i >= 0;) {
            *d++ = *buffer++;
        }
    } else {
        buffer += N - 1;
        for (int16_t i = N; --i >= 0;) {
            *d++ = *buffer--;
        }
    }
}

#endif /* SERIALIZER_H_ */
