/*
 Copyright (C) 2013 Stefano Zanotti.

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TIME_UTILS_H_
#define TIME_UTILS_H_

namespace FlashMat {

void nop() __attribute__((always_inline));

void delayNs(uint16_t ns) __attribute__((always_inline));
void delayUs(uint16_t us) __attribute__((always_inline));
void delayMs(uint16_t ms);

inline
void nop() {
    __asm__ __volatile__ ("nop");
}

// NOTE depending on the clock frequency, you can only get some amounts of delay:
// + 64MHz  -> 0ns ~ 65.500us  with resolution    62.5ns
// + 32MHz  -> 0ns ~ 65.500us  with resolution   125ns
// + 16MHz  -> 0ns ~ 65.500us  with resolution   250ns
// +  8MHz  -> 0ns ~ 65.500us  with resolution   500ns
// +  4MHz  -> 0ns ~ 65.000us  with resolution  1000ns
// +  2MHz  -> 0ns ~ 64.000us  with resolution  2000ns
// +  1MHz  -> 0ns ~ 64.000us  with resolution  4000ns
// Actual delays may be rounded to floor.
//
// NOTE: the actual timings rely on this function call being optimized, so that
// the manipulation of 'ns' is done at compile time.
// This might not be true if you provide a variable as argument; it's better
// to provide a constant (after all, this function is made for small and tweaked
// delays, so they should be well known at compile time).
inline
void delayNs(uint16_t ns) {

    uint16_t ns2;

    // all these shifts are optimized away
#if F_CPU == 64000000UL
    ns2 = ns / 125 * 2;
    ns  = (ns - ns2 * 125 / 2) * 4 / 125 * 2;
#elif F_CPU == 32000000UL
    ns2 = ns / 125;
    ns  = (ns - ns2 * 125) * 4 / 125;
#elif F_CPU == 16000000UL
    ns2 = ns / 250;
    ns  = (ns - ns2 * 250) * 2 / 125;
#elif F_CPU == 8000000UL
    ns2 = ns / 500;
    ns  = (ns - ns2 * 500) / 125;
#elif F_CPU == 4000000UL
    ns2 = ns / 1000;
    ns  = (ns - ns2 * 1000) / 250;
#elif F_CPU == 2000000UL
    ns2 = ns / 2000;
    ns  = (ns - ns2 * 2000) / 500;
#elif F_CPU == 1000000UL
    ns2 = ns / 4000;
    ns  = (ns - ns2 * 4000) / 1000;
#else
    #error "Clock frequency not supported yet"
#endif
        
    if (ns2 == 0 && ns == 0) { // (optimized away)
        return;
    }
    
    // now 'ns2' is actually the number of 4-cycle iterations.
    // We must remove 3 cycles (used to load the initial value for the loop: LDI LDI MOVW)
    // We must add 1 cycle (since the last brne executes in 1 cycle instead of 2,
    // since no branching is done).
    // All this only if 'ns2' > 0, otherwise no loop at all exists
    
    if(ns2 > 1) {
        // +3 loop init
        // -1 last branch
        ns2--;   // -4
        ns += 2; // +2
        // total = 0 (ok, no bias)
        
        if(ns >= 4) {
            ns -= 4;
            ns2++;
        }
    } else if (ns2 == 1) {
        // remove loop completely: no overhead to be taken into account
        ns2 = 0; // -4
        ns += 4; // +4
        // total = 0 (ok, no bias)
        
        if(ns >= 6) {
            ns2 = 1;
            ns -= 6;
        }
    } else {
        // ns is always < 4: never need to compact it into ns2
    }
    
    // from the calculations above, ns can never be > 5
    if (ns == 5) {
        nop(); nop(); nop(); nop(); nop();
    } else if (ns == 4) {
        nop(); nop(); nop(); nop();
    } else if (ns == 3) {
        nop(); nop(); nop();
    } else if (ns == 2) {
        nop(); nop();
    } else if (ns == 1) {
        nop();
    }
    
    if(ns2 == 0) { // (optimized away)
        return;
    }
    
    // busy wait
    __asm__ __volatile__ (
    	"1: sbiw %0,1" "\n\t" // 2 cycles
    	"brne 1b" : "=w" (ns2) : "0" (ns2) // 2 cycles (actually 1 when the loop ends)
    	);
}

// NOTE depending on the clock frequency, you can only get some amounts of delay:
// + 64MHz  -> 0us ~ 4095us   with "precision" 1/16us
// + 32MHz  -> 0us ~ 8191us   with "precision" 1/8us
// + 16MHz  -> 0us ~ 16383us  with "precision" 1/4us
// +  8MHz  -> 0us ~ 32767us  with "precision" 1/2us
// +  4MHz  -> 0us ~ 65535us  with resolution  1us
// +  2MHz  -> 0us ~ 65535us  with resolution  2us
// +  1MHz  -> 0us ~ 65535us  with resolution  4us
// ("precision" is the hardware precision, but since this function takes integers,
//  you cannot get it).
// Actual delays may be rounded to floor.
//
// NOTE: the actual timings rely on this function call being optimized, so that
// the manipulation of 'us' is done at compile time.
// This might not be true if you provide a variable as argument; it's better
// to provide a constant (after all, this function is made for small and tweaked
// delays, so they should be well known at compile time).
inline
void delayUs(uint16_t us) {

    // all these shifts are optimized away
#if F_CPU == 64000000UL
    us <<= 4;
#elif F_CPU == 32000000UL
    us <<= 3;
#elif F_CPU == 16000000UL
    us <<= 2;
#elif F_CPU == 8000000UL
    us <<= 1;
#elif F_CPU == 4000000UL
    // do nothing
#elif F_CPU == 2000000UL
    us >>= 1;
#elif F_CPU == 1000000UL
    us >>= 2;
#else
    #error "Clock frequency not supported yet"
#endif
    
    if (us == 0) { // (optimized away)
        return;
    }
    
    // now 'us' is actually the number of 4-cycle iterations
    // remove one cycle and add two nops:
    // the nops and the two LDI resulting from the constant argument
    // compensate the missing loop.
    // Since the brne below takes 1 cycle when it does not branch (exit condition),
    // we add a third nop here.
    
    us--; // remove one cycle (optimized away)
    nop(); nop(); nop();
    
    // busy wait
    __asm__ __volatile__ (
    	"1: sbiw %0,1" "\n\t" // 2 cycles
    	"brne 1b" : "=w" (us) : "0" (us) // 2 cycles (actually 1 when the loop ends)
    	);
}

inline
void delayMs(uint16_t ms) {
    // Does not compensate the overhead of the loops and decrements, but they
    // are quite small compared to milliseconds (if you call this function with a constant and
    // it gets inlined, they should be about 2/3 cycles to load the parameters and 4 cycles
    // for each iteration: given a delay of N ms, the additional delay is  ~= N*cpu_period;
    // even with 1MHz cpu (the lower the worse), we get a 0.4% error).
    
    while(ms >= 4) {
        delayUs(4000); // delayUs accepts max 4095 as argument, for some CPU frequencies
        ms -= 4;
    }
    
    while(ms) {
        delayUs(1000);
        ms--;
    }
}

} // namespace

#endif /* TIME_UTILS_H_ */